import unittest

import fastrlapi.core as core
import fastrlapi.env
import fastrlapi.backends.monitor as monitor
import fastrlapi.backends.debug as debug
import gym


class MonitorTest(unittest.TestCase):

    def setUp(self):
        self.env_name = fastrlapi.env._StepCountEnv.register_with_gym()
        if self.env_name in monitor._MonitorEnv._monitor_total_counts:
            monitor._MonitorEnv._monitor_total_counts[self.env_name] = monitor._MonitorTotalCounts(self.env_name)
        self.total = monitor._register_gym_monitor(self.env_name)
        self.env: monitor._MonitorEnv = monitor._get(gym.make(self.total.gym_env_name))
        self.env.reset()

    def tearDown(self):
        monitor._MonitorEnv._register_backend_agent(None)
        monitor._MonitorEnv._monitor_total_counts[self.env_name] = monitor._MonitorTotalCounts(self.env_name)

    def test_create(self):
        assert self.env
        assert self.env.total.instances_created == 1
        assert self.env.gym_env_name == self.env_name

    def test_max_steps_per_episode(self):
        self.env.max_steps_per_episode = 3
        while True:
            (state, reward, done, info) = self.env.step(1)
            if done:
                break
            if self.env.steps_done_in_episode > 1000:
                break
        assert self.env.steps_done_in_episode == 3

    def test_register_gym_monitor(self):
        assert self.env_name == self.total._original_env_name
        assert self.total.gym_env_name == monitor._MonitorEnv._NAME_PREFIX + self.env_name
        assert self.total is monitor._MonitorEnv._monitor_total_counts[self.env_name]

    def test_reset_beforeSteps_noEpisodeInc(self):
        self.env.reset()
        self.env.reset()
        assert self.total.episodes_done == 0
        assert self.env.episodes_done == 0

    def test_setbackendagent_twice(self):
        model_config = core.ModelConfig(self.env_name)
        agent = debug.DebugAgent(model_config)
        monitor._MonitorEnv._register_backend_agent(agent)
        monitor._MonitorEnv._register_backend_agent(agent)
        monitor._MonitorEnv._register_backend_agent(None)

    def test_setUp(self):
        assert self.env_name
        assert self.total
        assert self.env


if __name__ == '__main__':
    unittest.main()



import pytest
import unittest
import logging
from typing import Optional, Type

import fastrlapi
from fastrlapi.agents import CemAgent, DqnAgent, DoubleDqnAgent, DuelingDqnAgent, PpoAgent, \
    RandomAgent, ReinforceAgent, SacAgent, EasyAgent
from fastrlapi import env, core, agents
from fastrlapi.callbacks import duration, log, plot
import fastrlapi.backends.core

_step_count_name = env._StepCountEnv.register_with_gym()
_line_world_name = env._LineWorldEnv.register_with_gym()
_cartpole_name = "CartPole-v0"
_mountaincart_continuous_name = "MountainCarContinuous-v0"
fastrlapi.agents.seed = 0


def get_backends(agent: Optional[Type[EasyAgent]] = None):
    result = [b for b in agents.get_backends(agent) if b != 'default']
    assert result, f'no backend found for agent {agent}.'
    return result


def max_avg_rewards(tc: core.TrainContext):
    result = max([avg_rewards for (min_rewards, avg_rewards, max_rewards) in tc.eval_rewards.values()])
    return result


# noinspection PyTypeChecker
class BackendRegistrationTest(unittest.TestCase):
    class MyBackend(fastrlapi.backends.core.BackendAgentFactory):
        backend_name = "MyBackend"

    def setUp(self):
        self._oldbackends = agents._backends.copy()

    def tearDown(self):
        agents._backends = self._oldbackends

    def test_getbackends(self):
        assert agents._backends is not None
        assert agents.get_backends() is not None

    def test_getbackends_ppoagent(self):
        assert agents._backends is not None
        backends = agents.get_backends(agents.PpoAgent)
        assert 'default' in backends
        assert 'tfagents' in backends

    def test_getbackends_randomagent(self):
        assert agents._backends is not None
        backends = agents.get_backends(agents.RandomAgent)
        assert 'default' in backends
        assert 'tfagents' in backends

    def test_prepare_callbacks(self):
        agent = agents.PpoAgent(_line_world_name)
        c = [plot.ToMovie(), plot.Rewards()]
        d = agent._add_plot_callbacks(c, default_plots=None, default_plot_callbacks=[])
        assert isinstance(d[0], plot._PreProcess)
        assert isinstance(d[1], plot.Rewards)
        assert isinstance(d[-2], plot._PostProcess)
        assert isinstance(d[-1], plot.ToMovie)

    def test_register(self):
        old_len = len(fastrlapi.agents.get_backends())
        agents.register_backend(BackendRegistrationTest.MyBackend())
        b = fastrlapi.agents.get_backends()
        assert (old_len + 1) == len(b)
        assert BackendRegistrationTest.MyBackend.backend_name in b

    def test_register_backend_twice(self):
        b2 = fastrlapi.backends.default.DefaultAgentFactory(register_tensorforce=False)
        old_length = len(fastrlapi.agents.get_backends())
        assert b2 not in agents._backends
        agents.register_backend(backend=b2)
        assert old_length == len(fastrlapi.agents.get_backends())
        assert b2 in agents._backends

    def test_register_backend_None_exception(self):
        with pytest.raises(AssertionError):
            agents.register_backend(backend=None)


class CemAgentTest(unittest.TestCase):

    def train_and_assert(self, agent_type, num_iterations=100):
        logger = logging.warning
        for backend in get_backends(agent_type):
            logger(f'backend={backend} agent={agent_type}, num_iterations={num_iterations}')
            cem_agent: CemAgent = agent_type('CartPole-v0', fc_layers=(100,), backend=backend)
            tc: core.TrainContext = cem_agent.train([log.Duration(), log.Iteration(eval_only=True), log.Agent()],
                                                    num_iterations=num_iterations,
                                                    num_iterations_between_eval=10,
                                                    max_steps_per_episode=200,
                                                    default_plots=False)
            (min_steps, avg_steps, max_steps) = tc.eval_steps[tc.episodes_done_in_training]
            assert max_steps >= 100
            assert avg_steps >= 50

    def test_create(self):
        with pytest.raises(Exception):
            CemAgent('LineWorld-v0', fc_layers=(100,))


class DqnAgentsTest(unittest.TestCase):

    def train_and_eval(self, agent_type, backend, num_iterations):
        dqn_agent: DqnAgent = agent_type(_cartpole_name, fc_layers=(100,), backend=backend)
        tc: core.TrainContext = dqn_agent.train([log.Duration(), log.Iteration(eval_only=True), log.Agent()],
                                                num_iterations=num_iterations,
                                                num_steps_buffer_preload=1000,
                                                num_iterations_between_eval=500,
                                                max_steps_per_episode=200,
                                                default_plots=False)
        return max_avg_rewards(tc)

    def train_and_assert(self, agent_type, num_iterations=1000):
        logger = logging.warning
        for backend in get_backends(agent_type):
            current_num_iterations = num_iterations
            if backend == 'tensorforce':
                current_num_iterations = num_iterations * 3
            logger(f'backend={backend} agent={agent_type}, num_iterations={current_num_iterations}')
            rewards = self.train_and_eval(agent_type=agent_type, backend=backend, num_iterations=current_num_iterations)
            assert rewards >= 20, f'agent_type={agent_type} backend={backend} num_iterations={num_iterations}'

    def test_dqn(self):
        agents._activate_tfagents()
        self.train_and_assert(DqnAgent)

    @pytest.mark.skipif(fastrlapi.backends.core._tf_eager_execution_active, reason="_tf_eager_execution_active")
    @pytest.mark.tforce
    def test_dqn_tforce(self):
        agents.activate_tensorforce()
        self.train_and_assert(DqnAgent)

    def test_double_dqn(self):
        agents._activate_tfagents()
        with pytest.raises(Exception):
            self.train_and_assert(DoubleDqnAgent)

    @pytest.mark.skipif(fastrlapi.backends.core._tf_eager_execution_active, reason="_tf_eager_execution_active")
    @pytest.mark.tforce
    def test_dueling_dqn_tforce(self):
        agents.activate_tensorforce()
        self.train_and_assert(DuelingDqnAgent)



class EasyAgentTest(unittest.TestCase):

    def assert_properties_for_metric(self, metric, num_episodes):
        assert metric.min <= metric.max
        assert metric.mean <= metric.max
        assert metric.mean >= metric.min
        assert metric.std >= 0
        assert len(metric.all) == num_episodes

    def test_agent_saver_set(self):
        random_agent = RandomAgent(_line_world_name)
        assert random_agent._backend_agent._agent_context._agent_saver

    def test_evaluate(self):
        random_agent = RandomAgent(_line_world_name)
        num_episodes = 5
        metrics = random_agent.evaluate(num_episodes=num_episodes)
        self.assert_properties_for_metric(metrics.steps, num_episodes)
        self.assert_properties_for_metric(metrics.rewards, num_episodes)

    def test_save_no_trained_policy_exception(self):
        p1 = PpoAgent(gym_env_name=_line_world_name, fc_layers=(10, 20, 30), backend='tfagents')
        with pytest.raises(Exception):
            p1.save()

    def test_save_load_play(self):
        oldseed = agents.seed
        agents.seed = 123
        p1 = PpoAgent(gym_env_name=_line_world_name, fc_layers=(10, 20, 30), backend='tfagents')
        p1.train(callbacks=[duration._SingleEpisode()], default_plots=False)
        d = p1.save()
        agents.seed = oldseed
        p2: EasyAgent = agents.load(d)
        self.assert_are_equal(p1, p2)
        p2.play(default_plots=False, num_episodes=1)

    def test_seed(self):
        oldseed = agents.seed
        agents.seed = 123
        random_agent = RandomAgent(_line_world_name)
        assert random_agent._model_config.seed == 123
        agents.seed = oldseed

    def test_to_dict_from_dict(self):
        oldseed = agents.seed
        agents.seed = 123
        p1 = PpoAgent(gym_env_name=_line_world_name, fc_layers=(10, 20, 30), backend='tfagents')
        d = p1._to_dict()
        agents.seed = oldseed
        p2: EasyAgent = EasyAgent._from_dict(d)
        self.assert_are_equal(p1, p2)

    def assert_are_equal(self, p1: EasyAgent, p2: EasyAgent):
        assert p1
        assert p2
        assert p2._backend_name == p1._backend_name
        assert p2._backend_agent.__class__ == p1._backend_agent.__class__
        assert p2._model_config.original_env_name == p1._model_config.original_env_name
        assert p2._model_config.seed == p1._model_config.seed
        assert p2._model_config.gym_env_name == p1._model_config.gym_env_name
        assert p2._model_config.fc_layers == p1._model_config.fc_layers


# noinspection PyTypeChecker
class PpoAgentTest(unittest.TestCase):

    def test_callback_single(self):
        for backend in get_backends(PpoAgent):
            env._StepCountEnv.clear()
            agent = PpoAgent(_step_count_name, backend=backend)
            agent.train(duration._SingleEpisode())
            assert env._StepCountEnv.reset_count <= 2

    def test_train(self):
        agents.seed = 0
        for backend in get_backends(PpoAgent):
            ppo = PpoAgent(gym_env_name=_cartpole_name, backend=backend)
            tc = core.PpoTrainContext()
            tc.num_iterations = 10
            tc.num_episodes_per_iteration = 10
            tc.max_steps_per_episode = 200
            tc.num_epochs_per_iteration = 5
            tc.num_iterations_between_eval = 5
            tc.num_episodes_per_eval = 5
            ppo.train([log.Iteration()], train_context=tc, default_plots=False)
            assert max_avg_rewards(tc) >= 50

    @pytest.mark.skipif(fastrlapi.backends.core._tf_eager_execution_active, reason="_tf_eager_execution_active")
    @pytest.mark.tforce
    def test_train_tforce(self):
        agents.activate_tensorforce()
        self.test_train()

    def test_train_single_episode(self):
        for backend in get_backends(PpoAgent):
            ppo = agents.PpoAgent(gym_env_name=_step_count_name, backend=backend)
            count = log._CallbackCounts()
            ppo.train([log.Agent(), count, duration._SingleEpisode()])
            assert count.gym_init_begin_count == count.gym_init_end_count == 1
            assert count.gym_step_begin_count == count.gym_step_end_count
            assert count.gym_step_begin_count < 10 + count.gym_reset_begin_count

    def test_play_single_episode(self):
        for backend in get_backends(PpoAgent):
            ppo = agents.PpoAgent(gym_env_name=_step_count_name, backend=backend)
            count = log._CallbackCounts()
            cb = [log.Agent(), count, duration._SingleEpisode()]
            ppo.train(duration._SingleEpisode())
            ppo.play(cb)
            assert count.gym_init_begin_count == count.gym_init_end_count == 1
            assert count.gym_step_begin_count == count.gym_step_end_count <= 10

    def test_save_load(self):
        for backend in get_backends(PpoAgent):
            ppo = agents.PpoAgent(gym_env_name=_step_count_name, backend=backend)
            ppo.train([duration._SingleEpisode()], default_plots=False)
            temp_dir = ppo.save()
            ppo = agents.load(temp_dir)
            ppo.play(default_plots=False, num_episodes=1, callbacks=[])
            fastrlapi.backends.core._rmpath(temp_dir)

    @pytest.mark.skipif(fastrlapi.backends.core._tf_eager_execution_active, reason="_tf_eager_execution_active")
    @pytest.mark.tforce
    def test_save_load_tforce(self):
        agents.activate_tensorforce()
        self.test_save_load()


class RandomAgentTest(unittest.TestCase):

    def test_train(self):
        for backend in get_backends(RandomAgent):
            random_agent = RandomAgent(_line_world_name, backend=backend)
            tc: core.TrainContext = random_agent.train([log.Duration(), log.Iteration()],
                                                       num_iterations=10,
                                                       max_steps_per_episode=100,
                                                       default_plots=False)
            r = max_avg_rewards(tc)
            assert r >= 0

    @pytest.mark.skipif(fastrlapi.backends.core._tf_eager_execution_active, reason="_tf_eager_execution_active")
    @pytest.mark.tforce
    def test_train_tforce(self):
        agents.activate_tensorforce()
        self.test_train()


class ReinforceAgentTest(unittest.TestCase):

    def test_train(self):
        for backend in get_backends(RandomAgent):
            reinforce_agent: ReinforceAgent = ReinforceAgent(_line_world_name, backend=backend)
            tc: core.TrainContext = reinforce_agent.train([log.Duration(), log.Iteration()],
                                                          num_iterations=10,
                                                          max_steps_per_episode=200,
                                                          default_plots=False)
            r = max_avg_rewards(tc)
            assert r >= 5

    @pytest.mark.skipif(fastrlapi.backends.core._tf_eager_execution_active, reason="_tf_eager_execution_active")
    @pytest.mark.tforce
    def test_train_tforce(self):
        agents.activate_tensorforce()
        self.test_train()


class SacAgentTest(unittest.TestCase):

    def test_train(self):
        for backend in get_backends(SacAgent):
            sac_agent: SacAgent = SacAgent(_mountaincart_continuous_name, backend=backend)
            tc: core.TrainContext = sac_agent.train([log.Duration(), log.Iteration(eval_only=True), duration.Fast()],
                                                    default_plots=False)
            r = max_avg_rewards(tc)
            assert r >= -1


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import tempfile

import fastrlapi
from fastrlapi import agents
from fastrlapi.callbacks import save, duration

_line_world_name = fastrlapi.env._LineWorldEnv.register_with_gym()


class SaveTest(unittest.TestCase):

    def test_directory(self):
        best = save.Best()
        assert best.directory
        temp_path = os.path.abspath(os.path.join(tempfile.gettempdir(), "save_test"))
        best = save.Best(temp_path)
        assert temp_path == best.directory

    def test_best(self):
        agent = agents.PpoAgent(_line_world_name)
        best = save.Best()
        agent.train([best],
                    num_iterations_between_eval=1, num_episodes_per_iteration=10, num_iterations=3,
                    default_plots=False)
        os.path.isdir(best.directory)
        assert len(best.saved_agents) > 0
        (episode,reward,dir) = best.saved_agents[0]
        os.path.isdir(dir)
        agent2 = fastrlapi.agents.load(dir)
        assert agent2
        agent2.evaluate()

    def test_every(self):
        agent = agents.PpoAgent(_line_world_name)
        every = save.Every(num_evals_between_save=1)
        agent.train([every],
                    num_iterations_between_eval=1, num_episodes_per_iteration=10, num_iterations=3,
                    default_plots=False)
        os.path.isdir(every.directory)
        assert len(every.saved_agents) == 4
        for (episode,reward,dir) in every.saved_agents:
            os.path.isdir(dir)


if __name__ == '__main__':
    unittest.main()



import unittest

from fastrlapi import env, core, agents
from fastrlapi.callbacks import duration, log

_step_count_name  = env._StepCountEnv.register_with_gym()

class LogCallbacksTest(unittest.TestCase):

    def test_log_callbacks(self):
        agent = agents.PpoAgent(_step_count_name)
        agent.train([log._Callbacks(), duration._SingleIteration()])

    def test_log_agent(self):
        agent = agents.PpoAgent(_step_count_name)
        agent.train([log.Agent(), duration.Fast()])

    def test_log_iteration(self):
        agent = agents.PpoAgent(_step_count_name)
        agent.train([log.Iteration(), duration._SingleIteration()])

    def test_log_step(self):
        agent = agents.PpoAgent(_step_count_name)
        agent.train([log.Step(), duration._SingleIteration()])

    def test_cartpole_log_iteration(self):
        ppo = agents.PpoAgent(gym_env_name="CartPole-v0", backend='tfagents')
        ppo.train([log.Iteration(), duration._SingleIteration()])




"""This module contains backend core classes like Backend and BackendAgent.

    The concrete backends like tfagent or baselines are implemented in seprate modules.
"""
from abc import ABC, ABCMeta, abstractmethod
from typing import List, Optional, Tuple, Type, Dict
import datetime
import gym
import os
import shutil
import tempfile
import tensorflow

import numpy
import random

from fastrlapi import core
from fastrlapi.backends import monitor
from fastrlapi.callbacks import plot

_tf_eager_execution_active : Optional[bool] = None

def _get_temp_path():
    """Yields a path to a non-existent temporary directory inside the systems temp path."""
    result = os.path.join(tempfile.gettempdir(), tempfile.gettempprefix())
    n = datetime.datetime.now()
    result = result + f'-{n.year % 100:2}{n.month:02}{n.day:02}-{n.hour:02}{n.minute:02}{n.second:02}-' + \
             f'{n.microsecond:06}'
    return result


def _mkdir(directory: str):
    """Creates a directory with the given path. NoOps if the directory already exists.

        If a file exists at path, the file is removed.

        Returns:
            the absolute path to directory
    """
    directory = os.path.abspath(directory)
    if os.path.isfile(directory):
        _rmpath(directory)
    os.makedirs(directory, exist_ok=True)
    return directory


def _rmpath(path: str):
    """Removes the file or directory and its content. NoOps if the directory does not exist.

    Errors are ignored.
    """
    if path:
        if os.path.isdir(path):
            shutil.rmtree(path, ignore_errors=True)
        if os.path.isfile(path):
            os.remove(path)


class _BackendEvalCallback(core.AgentCallback):
    """Evaluates an agents current policy and updates its train_context accordingly."""

    def __init__(self, train_context: core.TrainContext):
        assert train_context, "train_context not set"
        assert train_context.num_episodes_per_eval > 0, "num_episodes_per_eval is 0."

        self._train_contex = train_context

    def on_play_episode_end(self, agent_context: core.AgentContext):
        pc = agent_context.play
        tc = self._train_contex
        sum_of_r = pc.sum_of_rewards.values()
        tc.eval_rewards[tc.episodes_done_in_training] = (
            min(sum_of_r), sum(sum_of_r) / len(sum_of_r), max(sum_of_r))
        steps = [len(episode_rewards) for episode_rewards in pc.rewards.values()]
        tc.eval_steps[tc.episodes_done_in_training] = (min(steps), sum(steps) / len(steps), max(steps))


class _BackendAgent(ABC):
    """Base class for all backend agent implementations.

        Implements the train loop and calls the Callbacks.
    """

    def __init__(self, model_config: core.ModelConfig, backend_name: str, tf_eager_execution: bool):
        """
        Args:
            model_config: defines the model and environment to be used
            backend_name: id of the backend to which this agent belongs to.
        """
        global _tf_eager_execution_active

        assert model_config is not None, "model_config not set."
        assert backend_name

        if _tf_eager_execution_active is None:
            _tf_eager_execution_active = tf_eager_execution
        assert _tf_eager_execution_active == tf_eager_execution, \
            "Due to an incompatibility between tensorforce and tfagents their agents can not be instantiated in the" +\
            "same python runtime instance (conflicting excpectations on tensorflows eager execution mode)."

        self._backend_name: str = backend_name
        self.model_config = model_config
        self._agent_context: core.AgentContext = core.AgentContext(self.model_config)
        self._agent_context.gym._totals = monitor._register_gym_monitor(self.model_config.original_env_name)
        self.model_config.gym_env_name = self._agent_context.gym._totals.gym_env_name

        self._preprocess_callbacks: List[core._PreProcessCallback] = [plot._PreProcess()]
        self._callbacks: List[core.AgentCallback] = []
        self._postprocess_callbacks: List[core._PostProcessCallback] = [plot._PostProcess()]

        self._train_total_episodes_on_iteration_begin: int = 0

    def _set_seed(self):
        """ sets the random seeds for all dependent packages """
        if not self.model_config.seed is None:
            seed = self.model_config.seed
            self.log_api(f'tf.random.set_seed', f'(seed={seed})')
            tensorflow.random.set_seed(seed=seed)
            self.log_api(f'numpy.random.seed', f'({seed})')
            numpy.random.seed(seed)
            self.log_api(f'random.seed', f'({seed})')
            random.seed(seed)
        return

    def _eval_current_policy(self):
        """Evaluates the current policy using play and updates the train_context

            If num_episodes_per_eval or num_iterations_per_eval is 0 no evaluation is performed.
        """
        tc = self._agent_context.train
        assert tc, "train_context not set"

        if tc.num_episodes_per_eval and tc.num_iterations_between_eval:
            callbacks = [_BackendEvalCallback(self._agent_context.train)] + self._callbacks
            self.play(play_context=core.PlayContext(self._agent_context.train), callbacks=callbacks)

    def log_api(self, api_target: str, log_msg: Optional[str] = None):
        """Logs a call to api_target with additional log_msg."""
        self._agent_context.gym._monitor_env = None
        if api_target is None:
            api_target = ''
        if log_msg is None:
            log_msg = ''
        for c in self._callbacks:
            c.on_api_log(self._agent_context, api_target, log_msg=log_msg)

    def log(self, log_msg: str):
        """Logs msg."""
        self._agent_context.gym._monitor_env = None
        if log_msg is None:
            log_msg = ''
        for c in self._callbacks:
            c.on_log(self._agent_context, log_msg=log_msg)

    def _on_gym_init_begin(self):
        """called when the monitored environment begins the instantiation of a new gym environment.

        Hint:
            the total instances count is not incremented yet."""
        self._agent_context.gym._monitor_env = None
        for c in self._callbacks:
            c.on_gym_init_begin(self._agent_context)
        self._agent_context.gym._monitor_env = None

    def _on_gym_init_end(self, env: monitor._MonitorEnv):
        """called when the monitored environment completed the instantiation of a new gym environment.

        Hint:
            o the total instances count is incremented by now
            o the new env (and its action space) is seeded with the api_context's seed
        """
        self._agent_context.gym._monitor_env = env
        if self._agent_context.model.seed is not None:
            env = self._agent_context.gym.gym_env
            seed = self._agent_context.model.seed
            env.seed(seed)
        for c in self._callbacks:
            c.on_gym_init_end(self._agent_context)
        self._agent_context.gym._monitor_env = None

    def _on_gym_reset_begin(self, env: monitor._MonitorEnv, **kwargs):
        """called when the monitored environment begins a reset.

        Hint:
            the total reset count is not incremented yet."""
        self._agent_context.gym._monitor_env = env
        for c in self._callbacks:
            c.on_gym_reset_begin(self._agent_context, **kwargs)
        self._agent_context.gym._monitor_env = None

    def _on_gym_reset_end(self, env: monitor._MonitorEnv, reset_result: Tuple, **kwargs):
        """called when the monitored environment completed a reset.

        Hint:
            the total episode count is incremented by now (if a step was performed before the last reset)."""
        self._agent_context.gym._monitor_env = env
        for c in self._callbacks:
            c.on_gym_reset_end(self._agent_context, reset_result, **kwargs)
        self._agent_context.gym._monitor_env = None

    def _on_gym_step_begin(self, env: monitor._MonitorEnv, action):
        """called when the monitored environment begins a step.

        Hint:
            o sets env.max_steps_per_episode if we are in train / play. Thus the episode is ended
              by the MonitorEnv if the step limit is exceeded
        """
        ac = self._agent_context
        ac.gym._monitor_env = env
        env.max_steps_per_episode = None
        if ac.is_play or ac.is_eval:
            env.max_steps_per_episode = ac.play.max_steps_per_episode
            self._on_play_step_begin(action)
        if ac.is_train:
            env.max_steps_per_episode = ac.train.max_steps_per_episode
            self._on_train_step_begin(action)
        for c in self._callbacks:
            c.on_gym_step_begin(self._agent_context, action)
        self._agent_context.gym._monitor_env = None

    def _on_gym_step_end(self, env: monitor._MonitorEnv, action, step_result: Tuple):
        """called when the monitored environment completed a step.

        Args:
            env: the gym_env the last step was done on
            step_result: the result (state, reward, done, info) of the last step call
        """
        ac = self._agent_context
        ac.gym._monitor_env = env
        if ac.is_play or ac.is_eval:
            self._on_play_step_end(action, step_result)
        if ac.is_train:
            self._on_train_step_end(action, step_result)
        for c in self._callbacks:
            c.on_gym_step_end(self._agent_context, action, step_result)
        self._agent_context.gym._monitor_env = None
        env.max_steps_per_episode = None

    def _on_play_begin(self):
        """Must NOT be called by play_implementation"""
        for c in self._callbacks:
            c.on_play_begin(self._agent_context)

    def _on_play_end(self):
        """Must NOT be called by play_implementation"""
        for c in self._callbacks:
            c.on_play_end(self._agent_context)
        self._agent_context.play.gym_env = None

    def on_play_episode_begin(self, env: gym.core.Env):
        """Must be called by play_implementation at the beginning of a new episode

        Args:
            env: the gym environment used to play the episode.
        """
        assert env, "env not set."
        assert isinstance(env, gym.core.Env), "env not an an instance of gym.Env."
        pc = self._agent_context.play
        pc.gym_env = env
        pc.steps_done_in_episode = 0
        pc.actions[pc.episodes_done + 1] = []
        pc.rewards[pc.episodes_done + 1] = []
        pc.sum_of_rewards[pc.episodes_done + 1] = 0

        for c in self._callbacks:
            c.on_play_episode_begin(self._agent_context)

    def on_play_episode_end(self):
        """Must be called by play_implementation at the end of an episode"""
        pc = self._agent_context.play
        pc.episodes_done += 1
        if pc.num_episodes and pc.episodes_done >= pc.num_episodes:
            pc.play_done = True

        for c in self._callbacks:
            c.on_play_episode_end(self._agent_context)

    def _on_play_step_begin(self, action):
        """Called before each call to gym.step on the current play env (agent_context.play.gym_env)

            Args:
                action: the action to be passed to the upcoming gym_env.step call
        """
        for c in self._callbacks:
            c.on_play_step_begin(self._agent_context, action)

    def _on_play_step_end(self, action, step_result: Tuple):
        """Called after each call to gym.step on the current play env (agent_context.play.gym_env)

        Args:
            step_result: the result (state, reward, done, info) of the last step call
        """
        (state, reward, done, info) = step_result
        pc = self._agent_context.play
        pc.steps_done_in_episode += 1
        pc.steps_done += 1
        pc.actions[pc.episodes_done + 1].append(action)
        pc.rewards[pc.episodes_done + 1].append(reward)
        pc.sum_of_rewards[pc.episodes_done + 1] += reward
        for c in self._callbacks:
            c.on_play_step_end(self._agent_context, action, step_result)

    def _on_train_begin(self):
        """Must NOT be called by train_implementation"""
        for c in self._callbacks:
            c.on_train_begin(self._agent_context)

    def _on_train_end(self):
        """Must NOT be called by train_implementation"""
        tc = self._agent_context.train
        if tc.episodes_done_in_training not in tc.eval_rewards:
            self._eval_current_policy()

        for c in self._callbacks:
            c.on_train_end(self._agent_context)

    def on_train_iteration_begin(self):
        """Must be called by train_implementation at the begining of a new iteration"""
        tc = self._agent_context.train
        tc.episodes_done_in_iteration = 0
        tc.steps_done_in_iteration = 0
        if tc.iterations_done_in_training == 0:
            self._eval_current_policy()
        self._train_total_episodes_on_iteration_begin = self._agent_context.gym._totals.episodes_done

        for c in self._callbacks:
            c.on_train_iteration_begin(self._agent_context)

    def on_train_iteration_end(self, loss: float, **kwargs):
        """Must be called by train_implementation at the end of an iteration

        Evaluates the current policy. Use kwargs to set additional dict values in train context.
        Eg for an ActorCriticTrainContext the losses may be set like this:
            on_train_iteration(loss=123,actor_loss=456,critic_loss=789)

        Args:
            loss: loss after the training of the model in this iteration or math.nan if the loss is not available
            **kwargs: if a keyword matches a dict property of the TrainContext instance, then
                        the dict[episodes_done_in_training] is set to the arg.
        """
        tc = self._agent_context.train
        totals = self._agent_context.gym._totals
        tc.episodes_done_in_iteration = (totals.episodes_done - self._train_total_episodes_on_iteration_begin)
        tc.episodes_done_in_training += tc.episodes_done_in_iteration
        tc.loss[tc.episodes_done_in_training] = loss

        # set traincontext dict from kwargs:
        for prop_name in kwargs:
            prop_instance = getattr(tc, prop_name, None)
            prop_value = kwargs[prop_name]
            if prop_instance is not None and isinstance(prop_instance, dict):
                prop_instance[tc.episodes_done_in_training] = prop_value

        tc.iterations_done_in_training += 1
        if tc.num_iterations is not None:
            tc.training_done = tc.iterations_done_in_training >= tc.num_iterations
        self._train_total_episodes_on_iteration_begin = 0
        if tc.num_iterations_between_eval and (tc.iterations_done_in_training % tc.num_iterations_between_eval == 0):
            self._eval_current_policy()

        for c in self._callbacks:
            c.on_train_iteration_end(self._agent_context)

    def _on_train_step_begin(self, action):
        """Called before each call to gym.step on the current train env (agent_context.train.gym_env)

            Args:
                action: the action to be passed to the upcoming gym_env.step call
        """
        pass

    def load(self, directory: str, callbacks: List[core.AgentCallback]):
        """Loads a previously trained and saved actor policy from directory.

        The loaded policy may afterwards be used by calling play().

        Args:
            directory: the directory containing the trained policy
            callbacks: list of callbacks called during the load.
        """
        assert callbacks is not None, "callbacks not set"
        assert directory
        assert os.path.isdir(directory)

        self._callbacks = callbacks
        self.load_implementation(directory)
        self._agent_context._is_policy_trained = True
        self._callbacks = None

    @abstractmethod
    def load_implementation(self, directory: str):
        """Loads a previously trained and saved actor policy from directory.

        The loaded policy may afterwards be used by calling play().

        Args:
            directory: the directory containing the trained policy.
        """

    # noinspection PyUnusedLocal
    def _on_train_step_end(self, action: object, step_result: Tuple):
        """Called after each call to gym.step on the current train env (agent_context.train.gym_env)

        Args:
            step_result: the result (state, reward, done, info) of the last step call
        """
        tc = self._agent_context.train
        tc.steps_done_in_iteration += 1
        tc.steps_done_in_training += 1

    def play(self, play_context: core.PlayContext, callbacks: List[core.AgentCallback]):
        """Forwarding to play_implementation overriden by the subclass.

            Args:
                play_context: play configuration to be used
                callbacks: list of callbacks called during play.
        """
        assert callbacks is not None, "callbacks not set"
        assert play_context, "play_context not set"
        assert self._agent_context.play is None, "play_context already set in agent_context"

        play_context._reset()
        play_context._validate()
        self._agent_context.play = play_context
        old_callbacks = self._callbacks
        self._callbacks = callbacks
        try:
            monitor._MonitorEnv._register_backend_agent(self)
            self._on_play_begin()
            self.play_implementation(self._agent_context.play)
            self._on_play_end()
        finally:
            monitor._MonitorEnv._register_backend_agent(None)
            self._callbacks = old_callbacks
            self._agent_context.play = None

    @abstractmethod
    def play_implementation(self, play_context: core.PlayContext):
        """Agent specific implementation of playing a single episode with the current policy.

            For implementation details see BackendBaseAgent.
        """

    def train(self, train_context: core.TrainContext, callbacks: List[core.AgentCallback]):
        """Forwarding to train_implementation overriden by the subclass

            Args:
                train_context: training configuration to be used
                callbacks: list of callbacks called during the training and evaluation.
        """
        assert callbacks is not None, "callbacks not set"
        assert train_context, "train_context not set"

        train_context._reset()
        train_context._validate()
        self._agent_context.train = train_context
        self._agent_context.play = None
        self._callbacks = callbacks

        try:
            self.log_api(f'backend_name', f'{self._backend_name}')
            self._set_seed()
            monitor._MonitorEnv._register_backend_agent(self)
            self._on_train_begin()
            self._agent_context._is_policy_trained = True
            self.train_implementation(self._agent_context.train)
            self._on_train_end()
        finally:
            monitor._MonitorEnv._register_backend_agent(None)
            self._callbacks = None
            self._agent_context.play = None
            self._agent_context.train = None

    @abstractmethod
    def train_implementation(self, train_context: core.TrainContext):
        """Agent specific implementation of the train loop.

            For implementational details see BackendBaseAgent.
        """

    def save(self, directory: str, callbacks: List[core.AgentCallback]):
        """Saves the currently trained actor policy in directory.

        Only the actor policy is guaranteed to be saved.
        Thus after a call to load resuming training is not supported.

        Args:
            directory: the directory to save the policy weights to. the directory must exist.
            callbacks: list of callbacks called during policy load.
        """
        assert callbacks is not None, "callbacks not set"
        assert directory
        assert os.path.isdir(directory)
        assert self._agent_context._is_policy_trained, "No trained policy available."

        self._callbacks = callbacks
        self.save_implementation(directory)
        self._callbacks = None

    @abstractmethod
    def save_implementation(self, directory: str):
        """Agent speecific implementation of saving the weights for the actor policy.

        Save must only guarantee to persist the weights of the actor policy.
        The implementation may write multiple files with fixed filenames.

        Args:
             directory: the (existing) directory to save the policy weights to.
        """


class BackendAgent(_BackendAgent, metaclass=ABCMeta):
    """Base class for all BackendAgent implementation.

        Explicitely exhibits all methods that should be overriden by an implementing agent.
    """

    @abstractmethod
    def load_implementation(self, directory: str):
        """Loads a previously trained and saved actor policy from directory.

        The loaded policy may afterwards be used by calling play().

        Args:
            directory: the directory containing the trained policy.
        """

    @abstractmethod
    def play_implementation(self, play_context: core.PlayContext):
        """Agent specific implementation of playing a number of episodes with the current policy.

            The implementation should have the form:

            while True:
                on_play_episode_begin(env)
                state = env.reset()
                while True:
                    action = _trained_policy.action(state)
                    (state, reward, done, info) = env.step(action)
                    if done:
                        break
                on_play_episode_end()
                if play_context.play_done:
                    break

            Args:
                play_context: play configuration to be used
        """

    @abstractmethod
    def save_implementation(self, directory: str):
        """Agent speecific implementation of saving the weights for the actor policy.

        Save must only guarantee to persist the weights of the actor policy.
        The implementation may write multiple files with fixed filenames.

        Args:
             directory: the directory to save the policy weights to.
        """

    @abstractmethod
    def train_implementation(self, train_context: core.TrainContext):
        """Agent specific implementation of the train loop.

            The implementation should have the form:

            while True:
                on_iteration_begin
                for e in num_episodes_per_iterations
                    play episode and record steps (while steps_in_episode < max_steps_per_episode and)
                train policy for num_epochs_per_iteration epochs
                on_iteration_end( loss )
                if training_done
                    break

            Args:
                train_context: context configuring the train loop

            Hints:
            o the subclasses training loss is passed through to BackendAgent by on_iteration_end.
              Thus the subclass must not add the experienced loss to the TrainContext.
        """


class BackendAgentFactory(ABC):
    """Backend agent factory defining the currently available agents (algorithms).
    """

    backend_name: str = 'abstract_BackendAgentFactory'

    def create_agent(self, easyagent_type: Type, model_config: core.ModelConfig) \
            -> Optional[_BackendAgent]:
        """Creates a backend agent instance implementing the algorithm given by agent_type.

        Args:
            easyagent_type: the EasyAgent derived type for which an implementing backend instance will be created
            model_config: the model_config passed to the constructor of the backend instance.

        Returns:
            instance of the agent or None if not implemented by this backend.
        """
        result: Optional[_BackendAgent] = None
        algorithms = self.get_algorithms()
        if easyagent_type in algorithms:
            result = algorithms[easyagent_type](model_config=model_config)
        return result

    def get_algorithms(self) -> Dict[Type, Type[_BackendAgent]]:
        """Yields a mapping of EasyAgent types to the implementations provided by this backend."""
        return {}



import unittest
import tempfile
import os
import math

from fastrlapi import agents, core
from fastrlapi.callbacks import duration, plot


class PlotTest(unittest.TestCase):

    def test_is_nan(self):
        p = plot._PlotCallback(core.PlotType.PLAY_STEP)
        assert not p._is_nan([])
        assert not p._is_nan(None)
        assert not p._is_nan([1])
        assert not p._is_nan([1,2])
        assert not p._is_nan([math.nan,1])
        assert p._is_nan([math.nan])
        assert p._is_nan([math.nan,math.nan])

    def test_play_plotstate(self):
        agent = agents.PpoAgent("CartPole-v0")
        agent.train([duration._SingleEpisode()])
        agent.play([plot.State()])

    def test_play_plotrewards(self):
        agent = agents.PpoAgent("CartPole-v0")
        agent.train([duration._SingleIteration()])
        agent.play([plot.Rewards()])

    def test_train_plotsteprewards(self):
        rndAgent = agents.RandomAgent('CartPole-v0')
        rndAgent.play([plot.StepRewards()] )

    def test_default_plots_False_nocallback(self):
        agent = agents.PpoAgent("CartPole-v0")
        p = plot.Loss()
        c = agent._add_plot_callbacks([], False, [p])
        assert not p in c

    def test_default_plots_None_durationcallback(self):
        agent = agents.PpoAgent("CartPole-v0")
        p = plot.Loss()
        c = agent._add_plot_callbacks([duration.Fast()], None, [p])
        assert p in c

    def test_default_plots_None_nocallback(self):
        agent = agents.PpoAgent("CartPole-v0")
        p = plot.Loss()
        c = agent._add_plot_callbacks([], None, [p])
        assert p in c

    def test_default_plots_None_plotcallback(self):
        agent = agents.PpoAgent("CartPole-v0")
        p = plot.Loss()
        r = plot.Rewards()
        c = agent._add_plot_callbacks([r], None, [p])
        assert not p in c
        assert r in c

    def test_default_plots_True_plotcallback(self):
        agent = agents.PpoAgent("CartPole-v0")
        p = plot.Loss()
        r = plot.Rewards()
        c = agent._add_plot_callbacks([r], True, [p])
        assert p in c
        assert r in c

    def test_train_plotsteprewards(self):
        ppoAgent = agents.PpoAgent('CartPole-v0')
        ppoAgent.train([plot.StepRewards(), duration.Fast()] )

    def test_train_plotloss(self):
        agent = agents.PpoAgent("CartPole-v0")
        agent.train([duration._SingleIteration(), plot.Loss()])

    def test_train_plotrewards(self):
        agent = agents.PpoAgent("CartPole-v0")
        agent.train([duration._SingleIteration(), plot.Rewards()])

    def test_train_plotstate(self):
        agent = agents.PpoAgent("CartPole-v0")
        agent.train([duration._SingleIteration(), plot.State()])

#    def test_train_plotstate_tensorforce(self):
#        agent = agents.PpoAgent("CartPole-v0", backend='tensorforce')
#        agent.train([duration.Fast(), plot.State()])

    def test_train_plotsteps(self):
        agent = agents.PpoAgent("CartPole-v0")
        agent.train([duration._SingleIteration(), plot.Steps()])

    def test_train_multiple_subplots(self):
        agent = agents.PpoAgent("CartPole-v0")
        agent.train([duration._SingleIteration(), plot.State(), plot.Rewards(), plot.Loss(), plot.Steps()])

    def test_train_tomovie(self):
        agent = agents.PpoAgent("CartPole-v0")
        agent.train([duration._SingleIteration(), plot.Rewards(), plot.ToMovie()])

    def test_train_tomovie_with_filename(self):
        f = tempfile.NamedTemporaryFile(delete=False)
        filepath = f.name
        f.close()
        os.remove(filepath)
        assert not os.path.isfile(filepath)
        filepath += '.gif'
        agent = agents.PpoAgent("CartPole-v0")
        m = plot.ToMovie(filepath=filepath, fps=10)
        agent.train([duration._SingleIteration(), plot.Rewards(), m])
        try:
            os.remove(m.filepath)
        except:
            pass

if __name__ == '__main__':
    unittest.main()



from typing import Optional, List, Tuple, Union, Dict

import fastrlapi.core as core
import fastrlapi.backends.core
import base64
import matplotlib.pyplot as plt
import numpy as np
import imageio
import math
import gym
import os.path

# avoid "double rendering" of the final jupyter output
on_play_end_clear_jupyter_display: bool = True
on_train_end_clear_jupyter_display: bool = True

# check if we are running in Jupyter, if so interactive plotting must be handled differently
# (in order to get plot updates during training)
_is_jupyter_active = False
try:
    # noinspection PyUnresolvedReferences
    from IPython import get_ipython
    # noinspection PyUnresolvedReferences
    from IPython.display import display, clear_output
    # noinspection PyUnresolvedReferences
    from IPython.display import HTML

    shell = get_ipython().__class__.__name__
    if shell == 'ZMQInteractiveShell':
        _is_jupyter_active = True
    else:
        # noinspection PyPackageRequirements
        import google.colab

        _is_jupyter_active = True
except ImportError:
    pass


class _PreProcess(core._PreProcessCallback):
    """Initializes the matplotlib agent_context.pyplot.figure"""

    def _setup(self, agent_context: core.AgentContext):
        # create figure / remove all existing axes from previous calls to train/play
        pyc = agent_context.pyplot
        pyc.is_jupyter_active = _is_jupyter_active

        if pyc.figure is None:
            pyc.figure = plt.figure("_fastrlapi", figsize=pyc.figsize)
        for ax in pyc.figure.axes:
            pyc.figure.delaxes(ax)

    def on_play_begin(self, agent_context: core.AgentContext):
        # play_begin is also called at the start of a policy evaluation
        if agent_context.is_play:
            self._setup(agent_context=agent_context)

    def on_train_begin(self, agent_context: core.AgentContext):
        self._setup(agent_context=agent_context)


class _PostProcess(core._PostProcessCallback):
    """Redraws the plots int the matplotlib agent_context.figure.

        In jupyter the plots are only refreshed once per second.
    """

    def __init__(self):
        self._call_jupyter_display: bool
        self._reset()

    def _clear_jupyter_plots(self, agent_context: core.AgentContext, wait=True):
        """Clears the content in the current jupyter output cell.
            NoOp if not in jupyter or no plot output.

        Args:
            wait: Wait to clear the output until new output is available to replace it.
        """
        # don't clear the jupyter output if no plot is present, may clear the log output otherwise
        if agent_context.pyplot.is_jupyter_active and self._plot_exists(agent_context):
            clear_output(wait=wait)

    def _display_plots(self, agent_context: core.AgentContext):
        """Fixes the layout of multiple subplots and refreshs the display."""
        pyc = agent_context.pyplot
        if self._plot_exists(agent_context):
            count = len(pyc.figure.axes)
            rows = math.ceil(count / pyc.max_columns)
            columns = math.ceil(count / rows)
            for i in range(count):
                pyc.figure.axes[i].change_geometry(rows, columns, i + 1)
            pyc.figure.tight_layout()

            if pyc.is_jupyter_active:
                self._clear_jupyter_plots(agent_context)
                if self._call_jupyter_display:
                    # noinspection PyTypeChecker
                    display(pyc.figure)
                self._call_jupyter_display = True
            else:
                plt.pause(0.01)

    def _plot_exists(self, agent_context: core.AgentContext):
        """Yields true if at least 1 jupyter plot exists."""
        pyc = agent_context.pyplot
        count = len(pyc.figure.axes)
        result = count > 0
        return result

    def _reset(self):
        self._call_jupyter_display = False

    def on_play_begin(self, agent_context: core.AgentContext):
        if agent_context.is_play:
            self._reset()

    def on_train_begin(self, agent_context: core.AgentContext):
        self._reset()

    def on_play_episode_end(self, agent_context: core.AgentContext):
        if agent_context._is_plot_ready(core.PlotType.PLAY_EPISODE):
            self._display_plots(agent_context)

    def on_play_step_end(self, agent_context: core.AgentContext, action, step_result: Tuple):
        if agent_context._is_plot_ready(core.PlotType.PLAY_STEP):
            self._display_plots(agent_context)

    def on_play_end(self, agent_context: core.AgentContext):
        if agent_context._is_plot_ready(core.PlotType.TRAIN_EVAL):
            self._display_plots(agent_context)
        if agent_context.is_play:
            self._display_plots(agent_context)
            if on_play_end_clear_jupyter_display:
                self._clear_jupyter_plots(agent_context, wait=False)

    def on_train_end(self, agent_context: core.AgentContext):
        self._display_plots(agent_context)
        if on_train_end_clear_jupyter_display:
            self._clear_jupyter_plots(agent_context, wait=False)

    def on_train_iteration_begin(self, agent_context: core.AgentContext):
        # display initial evaluation before training starts.
        if agent_context.train.iterations_done_in_training == 0 and \
                agent_context._is_plot_ready(core.PlotType.TRAIN_EVAL):
            self._display_plots(agent_context)

    def on_train_iteration_end(self, agent_context: core.AgentContext):
        if agent_context._is_plot_ready(core.PlotType.TRAIN_ITERATION):
            self._display_plots(agent_context)


# noinspection DuplicatedCode
class _PlotCallback(core.AgentCallback):
    """Base class of plyplot callbacks generating a plot after a trained iteration or an episode played.

        Attributes:
            axes: the subplot to plot onto
    """

    def __init__(self, plot_type: core.PlotType):
        """Base class of plyplot callbacks generating a plot after a trained iteration or an episode played.

            Args:
                plot_type: point in time when the plot is updated
        """
        self.axes = None
        self.axes_color: str = 'grey'
        self._plot_type: core.PlotType = plot_type

    def _create_subplot(self, agent_context: core.AgentContext):
        if self.axes is None:
            pyc = agent_context.pyplot
            pyc._created_subplots = pyc._created_subplots | self._plot_type
            count = len(pyc.figure.axes) + 1
            rows = math.ceil(count / pyc.max_columns)
            columns = math.ceil(count / rows)
            self.axes = pyc.figure.add_subplot(rows, columns, count)
            self.plot_axes(xlim=(0, 1), ylabel='', xlabel='')

    def _is_nan(self, values: Optional[List[float]]):
        """yields true if all values are equal to nan. yields false if values is None or empty."""
        result = False
        if values and all(isinstance(v, float) for v in values):
            result = all(math.isnan(v) for v in values)
        return result

    def _refresh_subplot(self, agent_context: core.AgentContext, plot_type: core.PlotType):
        """Sets this axes active and calls plot if this plot callback is registered on at least 1 plot
            out of plot_type."""
        assert self.axes is not None
        plot_type = plot_type & self._plot_type
        if agent_context._is_plot_ready(plot_type):
            pyc = agent_context.pyplot
            if not pyc.is_jupyter_active:
                plt.figure(pyc.figure.number)
                if plt.gcf() is pyc.figure:
                    plt.sca(self.axes)
            self.plot(agent_context, plot_type)

    def clear_plot(self, agent_context: core.AgentContext):
        """Clears the axes for this plot. Should be called by self.plot before replotting an axes."""
        assert self.axes is not None
        pyc = agent_context.pyplot
        if pyc.is_jupyter_active:
            self.axes.cla()
        else:
            plt.cla()

    def on_play_begin(self, agent_context: core.AgentContext):
        if (self._plot_type & (core.PlotType.PLAY_EPISODE | core.PlotType.PLAY_STEP)) != core.PlotType.NONE:
            self._create_subplot(agent_context)

    def on_play_end(self, agent_context: core.AgentContext):
        self._refresh_subplot(agent_context, core.PlotType.TRAIN_EVAL)

    def on_play_episode_end(self, agent_context: core.AgentContext):
        self._refresh_subplot(agent_context, core.PlotType.PLAY_EPISODE)

    def on_play_step_end(self, agent_context: core.AgentContext, action, step_result: Tuple):
        self._refresh_subplot(agent_context, core.PlotType.PLAY_STEP)

    def on_train_begin(self, agent_context: core.AgentContext):
        if (self._plot_type & (core.PlotType.TRAIN_EVAL | core.PlotType.TRAIN_ITERATION)) != core.PlotType.NONE:
            self._create_subplot(agent_context)

    def on_train_iteration_end(self, agent_context: core.AgentContext):
        self._refresh_subplot(agent_context, core.PlotType.TRAIN_ITERATION)

    def plot(self, agent_context: core.AgentContext, plot_type: core.PlotType):
        """Plots a graph on the self.axes object.

        Args:
            agent_context: the context containing the data to plot
            plot_type: the PlotType requesting this plot call.
        """
        pass

    def plot_axes(self, xlabel: str, ylabel: str, xlim: Tuple[float, float] = None,
                  yscale: str = 'linear', ylim: Optional[Tuple[float, float]] = None):
        """Draws the x- and y-axes.

        Attributes:
            xlim: None or (min,max) for the x-axes
            xlabel: label of the x-axes
            ylim: None or (min,max) for the y-axes (or None)
            ylabel: label of the y-axes
            yscale: scale of the y-axes ('linear', 'log', ...)
        """
        assert xlim is None or xlim[0] <= xlim[1]
        assert ylim is None or ylim[0] <= ylim[1]
        assert xlabel is not None
        assert ylabel is not None

        # setup subplot (axes, labels, colors)
        axes_color = self.axes_color
        self.axes.set_xlabel(xlabel)
        self.axes.set_ylabel(ylabel)
        self.axes.spines['top'].set_visible(False)
        self.axes.spines['right'].set_visible(False)
        self.axes.spines['bottom'].set_color(axes_color)
        self.axes.spines['left'].set_color(axes_color)
        self.axes.grid(color=axes_color, linestyle='-', linewidth=0.25, alpha=0.5)
        if xlim is not None:
            self.axes.set_xlim(xlim)
        if ylim is not None:
            self.axes.set_ylim(ylim)
        self.axes.set_yscale(yscale)

    def plot_text(self, text: str):
        if text:
            ax = self.axes
            ax.text(0.5, 0.5, text, horizontalalignment='center', verticalalignment='center',
                    color='blue', wrap=True)
            ax.set_xlabel('')
            ax.get_xaxis().set_ticks([])
            ax.get_yaxis().set_ticks([])
            axes_color = self.axes_color
            for spin in ax.spines:
                ax.spines[spin].set_visible(True)
                ax.spines[spin].set_color(axes_color)

    def plot_subplot(self, agent_context: core.AgentContext,
                     xvalues: List[int], yvalues: List[Union[float, Tuple[float, float, float]]], ylabel: str,
                     xlabel: str = None, xlim: Optional[Tuple[float, float]] = None,
                     yscale: str = 'linear', ylim: Optional[Tuple[float, float]] = None,
                     color: str = 'blue'):
        """Draws the graph given by xvalues, yvalues (including x- & y-axes) .

        Attributes:
            agent_context: context containing the figure to plot to
            xvalues: the graphs x-values (must have same length as y-values)
            yvalues: the graphs y-values or (min,y,max)-tuples (must have same length as x-values)
            ylim: (min,max) for the x-axes
            xlabel: label of the x-axes
            ylabel: label of the y-axes
            yscale: scale of the y-axes ('linear', 'log',...)
            ylim: (min,max) for the y-axes (or None)
            color: the graphs color (must be the name of a matplotlib color)
        """
        if not xlim:
            xmin = 0
            xmax = 1
            if agent_context.is_play:
                xmin = 1
                xmax = agent_context.play.episodes_done
            if agent_context.is_train or agent_context.is_eval:
                xmin = 0
                xmax = agent_context.train.episodes_done_in_training
            xlim = (xmin, xmax)
        if xlabel is None:
            xlabel = 'episodes'
            if agent_context.is_play:
                xlabel = 'episodes played'
            if agent_context.is_eval or agent_context.is_train:
                xlabel = 'episodes trained'
        self.clear_plot(agent_context)
        self.plot_axes(xlim=xlim, xlabel=xlabel, ylabel=ylabel, yscale=yscale, ylim=ylim)
        self.plot_values(agent_context=agent_context, xvalues=xvalues, yvalues=yvalues, color=color)

    def plot_values(self, agent_context: core.AgentContext,
                    xvalues: List[int], yvalues: List[Union[float, Tuple[float, float, float]]],
                    color: Optional[str] = 'blue', marker: str = None, pause: bool = True):
        """Draws the graph given by xvalues, yvalues.

        Attributes:
            agent_context: context containing the figure to plot to
            xvalues: the graphs x-values (must have same length as y-values)
            yvalues: the graphs y-values or (min,y,max)-tuples (must have same length as x-values)
            color: the graphs color (must be the name of a matplotlib color) or None
            marker: if None a marker is used if the graph has just a few points.
            pause: pause to redraw the plot.
        """
        assert xvalues is not None
        assert yvalues is not None
        assert len(xvalues) == len(yvalues), "xvalues do not match yvalues"

        pyc: core.PyPlotContext = agent_context.pyplot

        # extract min / max and y values if yvalues is of the form [(min,y,max),...)
        yminvalues = None
        ymaxvalues = None
        if len(yvalues) > 0 and isinstance(yvalues[0], tuple):
            ymaxvalues = [t[2] for t in yvalues]
            yminvalues = [t[0] for t in yvalues]
            yvalues = [t[1] for t in yvalues]

        # plot values
        marker = marker
        if marker is None and len(xvalues) < 10:
            marker = 'o'
        fill_alpha = 0.1
        if pyc.is_jupyter_active:
            if yminvalues is not None:
                self.axes.fill_between(xvalues, yminvalues, ymaxvalues, color=color, alpha=fill_alpha)
            self.axes.plot(xvalues, yvalues, color=color, marker=marker)
        else:
            if yminvalues is not None:
                plt.fill_between(xvalues, yminvalues, ymaxvalues, color=color, alpha=fill_alpha)
            plt.plot(xvalues, yvalues, color=color, marker=marker)
            if pause:
                plt.pause(0.01)


# noinspection DuplicatedCode
class Actions(_PlotCallback):

    def __init__(self, num_steps_between_plot=100):
        """Plots a histogram of the actions taken during play or during the last evaluation period.

        Args:
            num_steps_between_plot: num of steps to play before plot is updated.
        """
        super().__init__(plot_type=core.PlotType.PLAY_STEP | core.PlotType.PLAY_EPISODE | core.PlotType.TRAIN_EVAL)
        assert num_steps_between_plot > 0
        self._actions: List[float] = []
        self._num_steps_between_plot = num_steps_between_plot

    def _reset(self):
        self._actions: List[float] = []

    def on_play_begin(self, agent_context: core.AgentContext):
        super().on_play_begin(agent_context)
        self._reset()

    def on_play_end(self, agent_context: core.AgentContext):
        super().on_play_end(agent_context)
        self._reset()

    def plot(self, agent_context: core.AgentContext, plot_type: core.PlotType):
        try:
            pc = agent_context.play
            is_plot = True
            xlabel = 'actions'
            if plot_type & core.PlotType.TRAIN_EVAL != core.PlotType.NONE:
                xlabel = 'actions taken during last evaluation period'
            if plot_type & core.PlotType.PLAY_STEP != core.PlotType.NONE:
                is_plot = (pc.steps_done_in_episode % self._num_steps_between_plot == 0)
            if is_plot:
                self.clear_plot(agent_context)
                self.plot_axes(xlabel=xlabel, ylabel='count')
                self.axes.hist(pc.actions.values())
        except:
            self.plot_text(f'Failed to create the actions histogram.\n')


class Clear(core.AgentCallback):
    """Configures the clearing of plots in the jupyter output cell after calls to train or play."""

    def __init__(self, on_play: bool = True, on_train: bool = True):
        """Define the cell clearing behaviour after agent.train and agent.play.

        Args:
            on_play: if set the output cell is cleared after agent.play if a plot exists
            on_train: if set the output cell is cleared after agent.train if a plot exists
            """
        self._on_play: bool = on_play
        self._on_train: bool = on_train
        on_play_end_clear_jupyter_display = self._on_play
        on_train_end_clear_jupyter_display = self._on_train

    def on_play_begin(self, agent_context: core.AgentContext):
        global on_play_end_clear_jupyter_display

        on_play_end_clear_jupyter_display = self._on_play

    def on_train_begin(self, agent_context: core.AgentContext):
        global on_train_end_clear_jupyter_display

        on_train_end_clear_jupyter_display = self._on_train


class Loss(_PlotCallback):

    def __init__(self, yscale: str = 'symlog', ylim: Optional[Tuple[float, float]] = None):
        """Plots the loss resulting from each iterations policy training.

        Hints:
        o for actro-critic agents the loss from training the actor- and critic-networks are plotted
            along with the total loss.

        Args:
            yscale: scale of the y-axes ('linear', 'symlog',...)
            ylim: (min,max) for the y-axes
        """
        super().__init__(plot_type=core.PlotType.TRAIN_ITERATION)
        self.ylim = ylim
        self.yscale = yscale

    def plot(self, agent_context: core.AgentContext, plot_type: core.PlotType):
        ac = agent_context
        tc = ac.train
        xvalues = list(tc.loss.keys())
        self.clear_plot(agent_context)
        lossvalues = list(tc.loss.values())
        if self._is_nan(lossvalues):
            self.plot_text('no loss data available')
        else:
            self.plot_axes(xlim=(0, tc.episodes_done_in_training), xlabel='episodes trained',
                           ylim=self.ylim, ylabel='loss', yscale=self.yscale)
            if isinstance(tc, core.PpoTrainContext):
                acc: core.PpoTrainContext = tc
                self.plot_values(agent_context=ac, xvalues=xvalues, yvalues=lossvalues, color='indigo', pause=False)
                self.plot_values(agent_context=ac, xvalues=xvalues, yvalues=list(acc.actor_loss.values()), color='g',
                                 pause=False)
                self.plot_values(agent_context=ac, xvalues=xvalues, yvalues=list(acc.critic_loss.values()), color='b')
                self.axes.legend(('total', 'actor', 'critic'))
            else:
                self.plot_values(agent_context=ac, xvalues=xvalues, yvalues=lossvalues, color='indigo')


class Rewards(_PlotCallback):

    def __init__(self, yscale: str = 'linear', ylim: Optional[Tuple[float, float]] = None):
        """Plots the sum of rewards observed during policy evaluation.

        Args:
            yscale: scale of the y-axes ('linear', 'symlog',...)
            ylim: (min,max) for the y-axes
        """
        super().__init__(core.PlotType.TRAIN_ITERATION | core.PlotType.TRAIN_EVAL | core.PlotType.PLAY_EPISODE)
        self.ylim = ylim
        self.yscale = yscale

    def plot(self, agent_context: core.AgentContext, plot_type: core.PlotType):
        xvalues = yvalues = []
        ylabel = 'sum of rewards'
        if agent_context.is_train or agent_context.is_eval:
            tc = agent_context.train
            xvalues = list(tc.eval_rewards.keys())
            yvalues = list(tc.eval_rewards.values())
            ylabel = 'Ø sum of rewards'
        if agent_context.is_play:
            pc = agent_context.play
            xvalues = list(pc.sum_of_rewards.keys())
            yvalues = list(pc.sum_of_rewards.values())
        if xvalues:
            self.plot_subplot(agent_context, color='green', ylim=self.ylim, yscale=self.yscale,
                              xvalues=xvalues, yvalues=yvalues, ylabel=ylabel)


class State(_PlotCallback):
    """Renders the gym state as a plot to the pyplot figure using gym.render('rgb_array').

        During training only the last state of the last game evaluation is plotted.
        During play all states are plotted.
    """

    def __init__(self, mode='rgb_array'):
        """
        Args:
            mode: the render mode passed to gym.render(), yielding an rgb_array
        """
        super().__init__(plot_type=core.PlotType.PLAY_STEP | core.PlotType.TRAIN_EVAL)
        self._render_mode = mode

    def _plot_rgb_array(self, agent_context: core.AgentContext, rgb_array: np.ndarray):
        """Renders rgb_array to the current subplot."""
        assert rgb_array is not None
        ax = self.axes
        xlabel = ''
        if agent_context.is_eval:
            xlabel = "'done state' of last evaluation episode"
        ax.imshow(rgb_array)
        ax.set_xlabel(xlabel)
        ax.get_xaxis().set_ticks([])
        ax.get_yaxis().set_ticks([])
        axes_color = self.axes_color
        for spin in ax.spines:
            ax.spines[spin].set_visible(True)
            ax.spines[spin].set_color(axes_color)

    # noinspection PyArgumentList,DuplicatedCode
    def _render_to_rgb_array(self, gym_env: gym.Env, mode: str) -> np.ndarray:
        """ calls gym_env.render(mode) and validates the return value to be a numpy rgb array
            throws an exception if not an rgb array

            Returns:
                numpy rgb array
        """
        result = gym_env.render(mode=mode)

        assert result is not None, f'gym_env.render(mode={mode}) yielded None'
        assert isinstance(result, np.ndarray), f'gym_env.render(mode={mode}) did not yield a numpy.ndarray.'
        assert result.min() >= 0, f'gym_env.render(mode={mode}) contains negative values => not an rgb array'
        assert result.max() <= 255, f'gym_env.render(mode={mode}) contains values > 255 => not an rgb array'
        assert len(result.shape) == 3, f'gym_env.render(mode={mode}) shape is not of the form (x,y,n)'
        assert result.shape[2] == 3 or result.shape[2] == 4, \
            f'gym_env.render(mode={mode}) shape is not of the form (x,y,3|4)'
        return result

    def plot(self, agent_context: core.AgentContext, plot_type: core.PlotType):
        self.clear_plot(agent_context)
        try:
            rgb_array: np.ndarray = self._render_to_rgb_array(agent_context.play.gym_env, self._render_mode)
            self._plot_rgb_array(agent_context, rgb_array)
        except:
            self.plot_text(f'gym.Env.render(mode="{self._render_mode}") failed:\n')


class Steps(_PlotCallback):

    def __init__(self, yscale: str = 'linear', ylim: Optional[Tuple[float, float]] = None):
        """Plots the step counts observed during policy evaluation.

        Args:
            yscale: scale of the y-axes ('linear','log')
            ylim: (min,max) for the y-axes
        """
        super().__init__(core.PlotType.TRAIN_ITERATION | core.PlotType.TRAIN_EVAL | core.PlotType.PLAY_EPISODE)
        self.ylim = ylim
        self.yscale = yscale

    def plot(self, agent_context: core.AgentContext, plot_type: core.PlotType):
        xvalues = yvalues = []
        ylabel = 'steps'
        if agent_context.is_train or agent_context.is_eval:
            tc = agent_context.train
            xvalues = list(tc.eval_steps.keys())
            yvalues = list(tc.eval_steps.values())
            ylabel = 'Ø steps'
        if agent_context.is_play:
            pc = agent_context.play
            xvalues = list(pc.actions.keys())
            yvalues = [len(pc.actions[episode]) for episode in pc.actions.keys()]
        self.plot_subplot(agent_context, color='blue', ylim=self.ylim, yscale=self.yscale,
                          xvalues=xvalues, yvalues=yvalues, ylabel=ylabel)


class StepRewards(_PlotCallback):

    def __init__(self, num_steps_between_plot=100):
        """Plots the sum of rewards up to the current step during play or at the end of an
            evaluation period.

        Args:
            num_steps_between_plot: num of steps to play before plot is updated.
        """
        super().__init__(core.PlotType.PLAY_STEP | core.PlotType.PLAY_EPISODE | core.PlotType.TRAIN_EVAL)
        assert num_steps_between_plot > 0
        self._xy_values: Dict[int, Tuple[List[int], List[float]]] = dict()
        self._xmax: int = 0
        self._num_steps_between_plot = num_steps_between_plot

    def _replot(self, agent_context: core.AgentContext):
        if self._xmax >= 1:
            self.clear_plot(agent_context)
            xlabel = 'steps played'
            if agent_context.is_eval:
                xlabel = 'steps taken during last evaluation period'
            self.plot_axes(xlim=(1, self._xmax), ylabel='sum of rewards', xlabel=xlabel)
            xy_values = list(self._xy_values.values())
            xlast, _ = xy_values[-1]
            for xvalues, yvalues in xy_values:
                pause = (xvalues == xlast)
                self.plot_values(agent_context=agent_context, xvalues=xvalues, yvalues=yvalues,
                                 color=None, marker='', pause=pause)

    def _reset(self):
        self._xy_values = dict()
        self._xmax = 0

    def on_play_begin(self, agent_context: core.AgentContext):
        super().on_play_begin(agent_context)
        self._reset()

    def on_play_end(self, agent_context: core.AgentContext):
        super().on_play_end(agent_context)
        self._reset()

    def plot(self, agent_context: core.AgentContext, plot_type: core.PlotType):
        pc = agent_context.play
        if plot_type & core.PlotType.PLAY_STEP != core.PlotType.NONE:
            episode = pc.episodes_done + 1
            if episode not in self._xy_values:
                self._xy_values[episode] = ([], [])
            xvalues, yvalues = self._xy_values[episode]
            xvalues.append(pc.steps_done_in_episode)
            yvalues.append(pc.sum_of_rewards[episode])
            if pc.steps_done_in_episode > self._xmax:
                self._xmax = pc.steps_done_in_episode
            if (episode == 1 and pc.steps_done == 1) or (pc.steps_done % self._num_steps_between_plot) == 0:
                self._replot(agent_context)

        if plot_type & core.PlotType.PLAY_EPISODE != core.PlotType.NONE:
            self._replot(agent_context)

        if plot_type & core.PlotType.TRAIN_EVAL != core.PlotType.NONE:
            self._xmax = max([len(step_rewards) for step_rewards in pc.rewards.values()])
            for episode in pc.rewards.keys():
                step_rewards = pc.rewards[episode]
                xvalues = list(range(1, len(step_rewards) + 1))
                yvalues = []
                for reward in step_rewards:
                    old_sum = yvalues[-1] if len(yvalues) > 0 else 0
                    yvalues.append(old_sum + reward)
                self._xy_values[episode] = (xvalues, yvalues)
            self._replot(agent_context)


class ToMovie(core._PostProcessCallback):
    """Plots the pyplot figure to an mp4 file

    Attributes:
        fps: frame per seconds
        filepath: the filepath of the mp4 file.
    """

    def __init__(self, fps: Optional[int] = None, filepath: str = None):
        """Writes the ploted graphs and images to the mp4 / gif file given by filepath.

        if filepath ends in '.gif' an animated gif is created.

        Args:
            fps: frames per second
            filepath: the filepath of the mp4 or gif file file. If None the file is written to a temp file.
        """
        super().__init__()
        self.fps = fps
        self._is_filepath_set = filepath is not None
        self.filepath = filepath
        if not self._is_filepath_set:
            self.filepath = fastrlapi.backends.core._get_temp_path()
        if (not self._is_animated_gif()) and (not self.filepath.lower().endswith('.mp4')):
            self.filepath = self.filepath + '.mp4'
        self._video = imageio.get_writer(self.filepath, fps=fps) if fps else imageio.get_writer(self.filepath)

    def _close(self, agent_context: core.AgentContext):
        """closes the mp4 file and displays it in jupyter cell (if in a jupyter notebook)"""
        self._video.close()
        self._video = None
        if agent_context.pyplot.is_jupyter_active:
            with open(self.filepath, 'rb') as f:
                video = f.read()
                b64 = base64.b64encode(video)
            if not self._is_filepath_set:
                os.remove(self.filepath)
            width = 640
            height = 480
            if self._is_animated_gif():
                result = '''
                <img src="data:image/gif;base64,{2}" alt="fastrlapi.plot" width={0}/>
                '''.format(width, height, b64.decode())
            else:
                result = '''
                <video width="{0}" height="{1}" controls>
                    <source src="data:video/mp4;base64,{2}" type="video/mp4">
                Your browser does not support the video tag.
                </video>'''.format(width, height, b64.decode())
            result = HTML(result)
            # noinspection PyTypeChecker
            clear_output(wait=True)
            # noinspection PyTypeChecker
            display(result)

    def _get_rgb_array(self, agent_context: core.AgentContext) -> np.ndarray:
        """Yields an rgb array representing the current content of the subplots."""
        pyc = agent_context.pyplot
        pyc.figure.canvas.draw()
        result = np.frombuffer(pyc.figure.canvas.tostring_rgb(), dtype='uint8')
        result = result.reshape(pyc.figure.canvas.get_width_height()[::-1] + (3,))
        return result

    def _is_animated_gif(self):
        return self.filepath.lower().endswith('.gif')

    def _write_figure_to_video(self, agent_context: core.AgentContext):
        """Appends the current pyplot figure to the video.

        if an exception occures no frame is added.
        """
        try:
            rgb_array = self._get_rgb_array(agent_context)
            self._video.append_data(rgb_array)
        except:
            pass

    def on_play_episode_end(self, agent_context: core.AgentContext):
        if agent_context._is_plot_ready(core.PlotType.PLAY_EPISODE | core.PlotType.TRAIN_EVAL):
            self._write_figure_to_video(agent_context)

    def on_play_step_end(self, agent_context: core.AgentContext, action, step_result: Tuple):
        if agent_context._is_plot_ready(core.PlotType.PLAY_STEP):
            self._write_figure_to_video(agent_context)

    def on_train_iteration_end(self, agent_context: core.AgentContext):
        if agent_context._is_plot_ready(core.PlotType.TRAIN_ITERATION):
            self._write_figure_to_video(agent_context)

    def on_play_end(self, agent_context: core.AgentContext):
        if agent_context._is_plot_ready(core.PlotType.PLAY_EPISODE):
            self._close(agent_context)

    def on_train_end(self, agent_context: core.AgentContext):
        self._close(agent_context)



"""This module contains support classes and methods to interact with OpenAI gym environments
    as well as a gym env implementation for unit tests (linewordl)

    see https://github.com/openai/gym
"""
import gym, gym.envs, gym.error, gym.spaces
import inspect
import math
import matplotlib as plt
import numpy as np
from typing import List, Optional, Tuple, Dict, Callable, Any


def _is_registered_with_gym(gym_env_name: str) -> bool:
    """Determines if a gym environment with the name id exists.
    
        Args:
            gym_env_name: gym id to test.
            
        Returns:
            True if it exists, false otherwise
    """
    result = False
    try:
        spec = gym.envs.registration.spec(gym_env_name)
        assert spec is not None
        result = True
    except gym.error.UnregisteredEnv:
        pass
    return result


# noinspection DuplicatedCode
def register_with_gym(gym_env_name: str, entry_point: type, max_episode_steps: int = 100000, **kwargs):
    """Registers the class entry_point in gym by the name gym_env_name allowing overriding registrations.

    Thus different implementations of the same class (and the same name) maybe registered consecutively.
    The latest registrated version is used for instantiation.
    This facilitates developing an environment in a jupyter notebook without haveing to
    reregister a modified class under a new name.

    limitation: the max_episode_steps value of the first registration holds for all registrations
        with the same gym_env_name

    Args:
        gym_env_name: the gym environment name to be used as argument with gym.make
        max_episode_steps: all episodes end latest after this number of steps
        entry_point: the class to be registed with gym id gym_env_name
        kwargs: the args passed to the entry_point constructor call
    """
    assert gym_env_name is not None, "None is not an admissible environment name"
    assert type(gym_env_name) is str, "gym_env_name is not a str"
    assert len(gym_env_name) > 0, "empty string is not an admissible environment name"
    assert inspect.isclass(entry_point), "entry_point not a class"
    assert issubclass(entry_point, gym.Env), "entry_point not a subclass of gym.Env"
    assert callable(entry_point), "entry_point not callable"

    if gym_env_name not in _ShimEnv._entry_points:
        gym.envs.registration.register(id=gym_env_name,
                                       entry_point=_ShimEnv,
                                       max_episode_steps=max_episode_steps,
                                       kwargs={_ShimEnv._KWARG_GYM_NAME: gym_env_name})
    _ShimEnv._entry_points[gym_env_name] = (entry_point, kwargs)


class _LineWorldEnv(gym.Env):
    """Simple environment for fast unittest, registered as 'LineWorld-v0'

        * an agent lives in a finite linear world of uneven elements
        * at each moment it is in a certain position
        * initial position is the middle
        * some positions gain rewards, some don't
        * rewards are between 0 and 15
        * agent can either move left or right
        * objective: maximize total reward = sum(rewards) + sum(steps)
        * Cost per step: -1
        * Done Condition: agent is at pos 0 or total reward <= -20
    """

    @staticmethod
    def register_with_gym():
        """Register this environment with gym and yields the gym environment name."""
        result = "LineWorld-v0"
        register_with_gym(result, _LineWorldEnv)
        return result

    def __init__(self, world: Optional[List[int]] = None):
        """Creates the lineword, size and rewards are given by the world arg.

        Args:
            world: list of rewards to collect in each position of the lineworld.
        """
        if world is None:
            world = [10, 0, 0, 5, 0, 2, 15]
        assert world, "world must not be None or empty."
        self.world: np.array = np.array(world)
        number_of_actions: int = 2
        self.action_space: gym.spaces.Discrete = gym.spaces.Discrete(number_of_actions)
        self.size_of_world: int = len(world)
        self.max_reward: int = max(world)
        self.min_reward: int = min(world)

        # the environment's current state is described by the position of the agent and the remaining rewards
        self.observation_size: int = 1 + self.size_of_world
        low: np.array = np.full(self.observation_size, self.min_reward)
        high: np.array = np.full(self.observation_size, self.max_reward)
        self.observation_space: gym.spaces.Box = gym.spaces.Box(low=low, high=high, dtype=np.float32)
        self.reward_range: Tuple[int, int] = (self.min_reward, self.max_reward)
        self.steps: int = 0
        self.done: bool = False
        self.pos: int = 0
        self._figure = None
        self.reset()

    def get_observation(self):
        return np.append([self.pos], self.remaining_rewards)

    def reset(self):
        self.total_reward = 0
        self.done = False
        self.pos = math.floor(len(self.world) / 2)
        self.steps = 0
        self.remaining_rewards = np.array(self.world, copy=True)
        return self.get_observation()

    def step(self, action):
        """perform action on this lineword.

        Args:
            action: 0 ==> move left, 1 ==> move right
        """
        if isinstance(action, np.ndarray):
            assert action.size == 1, "action of type numpy.array as invalid size"
            action = (int)(action)
        if isinstance(action, np.int32):
            action = (int)(action)
        assert isinstance(action, int)
        if action <= 0 and self.pos > 0:
            self.pos -= 1
        if action > 0 and self.pos < self.size_of_world - 1:
            self.pos += 1

        reward = self.remaining_rewards[self.pos] - 1
        self.total_reward += reward
        self.remaining_rewards[self.pos] = 0

        self.done = (self.pos == 0) or (self.total_reward <= -20)
        self.steps += 1

        observation = self.get_observation()
        info = None
        return observation, reward, self.done, info

    def _render_to_ansi(self):
        return f'position: {self.pos}, remaining rewards: {self.remaining_rewards},' + \
               f'total reward so far: {self.total_reward}, steps so far: {self.steps}, game done: {self.done}'

    def _render_to_figure(self):
        """ Renders the current state as a graph with matplotlib """
        if self._figure is not None:
            plt.close(self._figure)
        self._figure, ax = plt.subplots(1, figsize=(8, 4))
        ax.set_ylim(bottom=-1, top=self.max_reward + 1)
        x = np.arange(0, self.size_of_world, 1, dtype=np.uint8)
        y = self.remaining_rewards
        plt.plot([self.pos, self.pos], [0, 2], 'r^-')
        ax.scatter(x, y, s=75)
        self._figure.canvas.draw()
        return self._figure

    def _render_to_rgb(self):
        """ convert the output of render_to_figure to a rgb_array """
        self._render_to_figure()
        self._figure.canvas.draw()
        buf = self._figure.canvas.tostring_rgb()
        num_cols, num_rows = self._figure.canvas.get_width_height()
        plt.close(self._figure)
        self._figure = None
        result = np.fromstring(buf, dtype=np.uint8).reshape(num_rows, num_cols, 3)
        return result

    def render(self, mode='ansi'):
        if mode == 'ansi':
            return self._render_to_ansi()
        elif mode == 'human':
            return self._render_to_figure()
        elif mode == 'rgb_array':
            return self._render_to_rgb()
        else:
            super().render(mode=mode)


class _ShimEnv(gym.Wrapper):
    """Wrapper to redirect the instantiation of a gym environment to its current implementation.
    """

    _KWARG_GYM_NAME = "shimenv_gym_name"
    _entry_points: Dict[str, Tuple[Callable, Dict]] = {}

    def __init__(self, **kwargs):
        assert _ShimEnv._KWARG_GYM_NAME in kwargs, f'{_ShimEnv._KWARG_GYM_NAME} missing from kwargs'

        self._gym_env_name = kwargs[_ShimEnv._KWARG_GYM_NAME]
        entry_point, gym_kwargs = _ShimEnv._entry_points[self._gym_env_name]
        self._gym_env = entry_point(**gym_kwargs)
        super().__init__(self._gym_env)

    def step(self, action):
        return self.env.step(action)

    def reset(self, **kwargs):
        return self.env.reset(**kwargs)


class _StepCountEnv(gym.core.Env):
    """Debug Env that runs forever, counting the calls to reset and step."""

    metadata = {'render.modes': ['ansi']}
    reward_range = (0, 1)
    max = 10 ** 7
    action_space = gym.spaces.discrete.Discrete(2)
    observation_space = gym.spaces.Box(low=0, high=max, shape=(1,))

    step_count: int = 0
    reset_count: int = 0

    @staticmethod
    def register_with_gym():
        """Register this environment with gym and yields the gym environment name."""
        result = "_StepCountEnv-v0"
        register_with_gym(result, _StepCountEnv)
        return result

    @staticmethod
    def clear():
        _StepCountEnv.step_count = 0
        _StepCountEnv.reset_count = 0

    def __str__(self):
        return f'reset_count={_StepCountEnv.reset_count} step_count={_StepCountEnv.step_count}'

    def step(self, action):
        _StepCountEnv.step_count += 1
        # noinspection PyRedundantParentheses
        return [_StepCountEnv.step_count], 1, False, None

    def reset(self):
        _StepCountEnv.reset_count += 1
        return [_StepCountEnv.step_count]

    def render(self, mode='ansi'):
        return str(self)



"""This module contains the backend implementation for tf Agents (see https://github.com/tensorflow/agents)"""
from abc import ABCMeta
from typing import Dict, Type
import math
import os

# noinspection PyUnresolvedReferences
import fastrlapi.agents
from fastrlapi import core
from fastrlapi.backends import core as bcore
from fastrlapi.backends import monitor

# noinspection PyPackageRequirements
import tensorflow as tf

from tf_agents.agents.ddpg import critic_network
from tf_agents.agents.dqn import dqn_agent
from tf_agents.agents.ppo import ppo_agent
from tf_agents.agents.reinforce import reinforce_agent
from tf_agents.agents.sac import sac_agent
from tf_agents.drivers import dynamic_step_driver
from tf_agents.drivers.dynamic_episode_driver import DynamicEpisodeDriver
from tf_agents.environments import gym_wrapper, py_environment, tf_py_environment
from tf_agents.networks import actor_distribution_network, normal_projection_network, q_network, value_network
from tf_agents.policies import greedy_policy, tf_policy, random_tf_policy, policy_saver
from tf_agents.replay_buffers import tf_uniform_replay_buffer
from tf_agents.replay_buffers.tf_uniform_replay_buffer import TFUniformReplayBuffer
from tf_agents.trajectories import trajectory
from tf_agents.utils import common

import gym


# noinspection PyUnresolvedReferences,PyAbstractClass
class TfAgent(bcore.BackendAgent, metaclass=ABCMeta):
    """Reinforcement learning agents based on googles tf_agent implementations

        https://github.com/tensorflow/agents
    """

    def __init__(self, model_config: core.ModelConfig):
        super().__init__(model_config=model_config,
                         backend_name=TfAgentAgentFactory.backend_name,
                         tf_eager_execution=True)
        self._trained_policy = None
        self._play_env: Optional[gym.Env] = None

    def _create_gym_with_wrapper(self, discount):
        gym_spec = gym.spec(self.model_config.gym_env_name)
        gym_env = gym_spec.make()

        # simplify_box_bounds: Whether to replace bounds of Box space that are arrays
        #  with identical values with one number and rely on broadcasting.
        # important, simplify_box_bounds True crashes environments with boundaries with identical values
        env = gym_wrapper.GymWrapper(
            gym_env,
            discount=discount,
            simplify_box_bounds=False)
        return env

    def _create_env(self, discount: float = 1) -> tf_py_environment.TFPyEnvironment:
        """ creates a new instance of the gym environment and wraps it in a tfagent TFPyEnvironment

            Args:
                discount: the reward discount factor
        """
        assert 0 < discount <= 1, "discount not admissible"

        self.log_api(f'TFPyEnvironment', f'( suite_gym.load( "{self.model_config.original_env_name}", discount={discount}) )')
        # suit_gym.load crashes our environment 
        # py_env = suite_gym.load(self.model_config.gym_env_name, discount=discount)
        py_env = self._create_gym_with_wrapper(discount)

        result = tf_py_environment.TFPyEnvironment(py_env)
        return result

    def _get_gym_env(self, tf_py_env: tf_py_environment.TFPyEnvironment) -> monitor._MonitorEnv:
        """ extracts the underlying _MonitorEnv from tf_py_env created by _create_tfagent_env"""
        assert isinstance(tf_py_env, tf_py_environment.TFPyEnvironment), \
            "passed tf_py_env is not an instance of TFPyEnvironment"
        assert isinstance(tf_py_env.pyenv, py_environment.PyEnvironment), \
            "passed TFPyEnvironment.pyenv does not contain a PyEnvironment"
        assert len(tf_py_env.pyenv.envs) == 1, "passed TFPyEnvironment.pyenv does not contain a unique environment"

        result = tf_py_env.pyenv.envs[0].gym
        assert isinstance(result, monitor._MonitorEnv), "passed TFPyEnvironment does not contain a _MonitorEnv"
        return result

    def play_implementation(self, play_context: core.PlayContext):
        """Agent specific implementation of playing a single episodes with the current policy.

            Args:
                play_context: play configuration to be used
        """
        assert play_context, "play_context not set."
        assert self._trained_policy, "trained_policy not set. call train() first."

        if self._play_env is None:
            self._play_env = self._create_env()
        gym_env = self._get_gym_env(self._play_env)
        while True:
            self.on_play_episode_begin(env=gym_env)
            time_step = self._play_env.reset()
            while not time_step.is_last():
                action_step = self._trained_policy.action(time_step)
                time_step = self._play_env.step(action_step.action)
            self.on_play_episode_end()
            if play_context.play_done:
                break


    def load_implementation(self, directory: str):
        """Loads a previously saved actor policy from the directory

        Args:
             directory: the directory to load the policy from.
        """
        assert directory

        self.log_api('saved_model.load', f'({directory})')
        self._trained_policy = tf.compat.v2.saved_model.load(directory)


    def save_implementation(self, directory: str):
        """Saves the trained actor policy in directory.
           If no policy was trained yet, no file is written
.
        Args:
             directory: the directory to save the policy weights to.
        """
        assert self._trained_policy, "no policy trained yet."

        self.log_api('PolicySaver', f'(trained_policy,seed={self.model_config.seed})')
        saver = policy_saver.PolicySaver(self._trained_policy, seed=self.model_config.seed)
        self.log_api('policy_saver.save', f'({directory})')
        saver.save(directory)


# noinspection PyUnresolvedReferences
class TfDqnAgent(TfAgent):
    """ creates a new agent based on the DQN algorithm using the tfagents implementation.

        Args:
            model_config: the model configuration including the name of the target gym environment
                as well as the neural network architecture.
    """

    def __init__(self, model_config: core.ModelConfig):
        super().__init__(model_config=model_config)

    def collect_step(self, env: tf_py_environment.TFPyEnvironment, policy: tf_policy.Base,
                     replay_buffer: TFUniformReplayBuffer):
        time_step = env.current_time_step()
        action_step = policy.action(time_step)
        next_time_step = env.step(action_step.action)
        traj = trajectory.from_transition(time_step, action_step, next_time_step)
        replay_buffer.add_batch(traj)

    # noinspection DuplicatedCode
    def train_implementation(self, train_context: core.TrainContext):
        """Tf-Agents Ppo Implementation of the train loop.

        The implementation follows
        https://colab.research.google.com/github/tensorflow/agents/blob/master/tf_agents/colabs/1_dqn_tutorial.ipynb
        """
        assert isinstance(train_context, core.StepsTrainContext)
        dc: core.StepsTrainContext = train_context

        train_env = self._create_env(discount=dc.reward_discount_gamma)
        observation_spec = train_env.observation_spec()
        action_spec = train_env.action_spec()
        timestep_spec = train_env.time_step_spec()

        # SetUp Optimizer, Networks and DqnAgent
        self.log_api('AdamOptimizer', f'(learning_rate={dc.learning_rate})')
        optimizer = tf.compat.v1.train.AdamOptimizer(learning_rate=dc.learning_rate)
        self.log_api('QNetwork', f'(observation_spec, action_spec, fc_layer_params={self.model_config.fc_layers})')
        q_net = q_network.QNetwork(observation_spec, action_spec, fc_layer_params=self.model_config.fc_layers)
        self.log_api('DqnAgent', '(timestep_spec,action_spec,q_network=..., optimizer=...,td_errors_loss_fn=common.element_wise_squared_loss)')
        tf_agent = dqn_agent.DqnAgent(timestep_spec, action_spec,
                                      q_network=q_net, optimizer=optimizer,
                                      td_errors_loss_fn=common.element_wise_squared_loss)

        self.log_api('tf_agent.initialize', f'()')
        tf_agent.initialize()
        self._trained_policy = tf_agent.policy

        # SetUp Data collection & Buffering
        self.log_api('TFUniformReplayBuffer', f'(data_spec=..., batch_size={train_env.batch_size}, max_length={dc.max_steps_in_buffer})')
        replay_buffer = TFUniformReplayBuffer(data_spec=tf_agent.collect_data_spec,
                                              batch_size=train_env.batch_size,
                                              max_length=dc.max_steps_in_buffer)
        self.log_api('RandomTFPolicy', '()')
        random_policy = random_tf_policy.RandomTFPolicy(timestep_spec, action_spec)
        self.log_api('replay_buffer.add_batch', '(trajectory)')
        for _ in range(dc.num_steps_buffer_preload):
            self.collect_step(env=train_env, policy=random_policy, replay_buffer=replay_buffer)

        # Train
        self.log_api('tf_agent.train', '= common.function(tf_agent.train)')
        tf_agent.train = common.function(tf_agent.train)
        self.log_api('replay_buffer.as_dataset', f'(num_parallel_calls=3, ' +
                     f'sample_batch_size={dc.num_steps_sampled_from_buffer}, num_steps=2).prefetch(3)')
        dataset = replay_buffer.as_dataset(num_parallel_calls=3, sample_batch_size=dc.num_steps_sampled_from_buffer,
                                           num_steps=2).prefetch(3)
        self.log_api('iter(dataset',f'{iter(dataset)}')
        iter_dataset = iter(dataset)
        self.log_api('for each iteration')
        self.log_api('  replay_buffer.add_batch', '(trajectory)')
        self.log_api('  tf_agent.train', '(experience=trajectory)')
        while True:
            self.on_train_iteration_begin()
            for _ in range(dc.num_steps_per_iteration):
                self.collect_step(env=train_env, policy=tf_agent.collect_policy, replay_buffer=replay_buffer)
            trajectories, _ = next(iter_dataset)
            tf_loss_info = tf_agent.train(experience=trajectories)
            self.on_train_iteration_end(tf_loss_info.loss)
            if train_context.training_done:
                break
        return


# noinspection PyUnresolvedReferences
class TfPpoAgent(TfAgent):
    """ creates a new agent based on the PPO algorithm using the tfagents implementation.
        PPO is an actor-critic algorithm using 2 neural networks. The actor network
        to predict the next action to be taken and the critic network to estimate
        the value of the game state we are currently in (the expected, discounted
        sum of future rewards when following the current actor network).

        Args:
            model_config: the model configuration including the name of the target gym environment
                as well as the neural network architecture.
    """

    def __init__(self, model_config: core.ModelConfig):
        super().__init__(model_config=model_config)

    # noinspection DuplicatedCode
    def train_implementation(self, train_context: core.TrainContext):
        """Tf-Agents Ppo Implementation of the train loop."""

        assert isinstance(train_context, core.PpoTrainContext)
        tc: core.PpoTrainContext = train_context
        train_env = self._create_env(discount=tc.reward_discount_gamma)
        observation_spec = train_env.observation_spec()
        action_spec = train_env.action_spec()
        timestep_spec = train_env.time_step_spec()

        # SetUp Optimizer, Networks and PpoAgent
        self.log_api('AdamOptimizer', '()')
        optimizer = tf.compat.v1.train.AdamOptimizer(learning_rate=tc.learning_rate)

        self.log_api('ActorDistributionNetwork', '()')
        actor_net = actor_distribution_network.ActorDistributionNetwork(observation_spec, action_spec,
                                                                        fc_layer_params=self.model_config.fc_layers)
        self.log_api('ValueNetwork', '()')
        value_net = value_network.ValueNetwork(observation_spec, fc_layer_params=self.model_config.fc_layers)

        self.log_api('PpoAgent', '()')
        tf_agent = ppo_agent.PPOAgent(timestep_spec, action_spec, optimizer,
                                      actor_net=actor_net, value_net=value_net,
                                      num_epochs=tc.num_epochs_per_iteration)
        self.log_api('tf_agent.initialize', '()')
        tf_agent.initialize()
        self._trained_policy = tf_agent.policy

        # SetUp Data collection & Buffering
        collect_data_spec = tf_agent.collect_data_spec
        self.log_api('TFUniformReplayBuffer', '()')
        replay_buffer = TFUniformReplayBuffer(collect_data_spec,
                                              batch_size=1, max_length=tc.max_steps_in_buffer)

        collect_policy = tf_agent.collect_policy
        self.log_api('DynamicEpisodeDriver', '()')
        collect_driver = DynamicEpisodeDriver(train_env, collect_policy, observers=[replay_buffer.add_batch],
                                              num_episodes=tc.num_episodes_per_iteration)

        # Train
        collect_driver.run = common.function(collect_driver.run, autograph=False)
        tf_agent.train = common.function(tf_agent.train, autograph=False)

        while True:
            self.on_train_iteration_begin()
            self.log_api('-----', f'iteration {tc.iterations_done_in_training:4} of {tc.num_iterations:<4}      -----')
            self.log_api('collect_driver.run', '()')
            collect_driver.run()

            self.log_api('replay_buffer.gather_all', '()')
            trajectories = replay_buffer.gather_all()

            self.log_api('tf_agent.train', '(experience=...)')
            loss_info = tf_agent.train(experience=trajectories)
            total_loss = loss_info.loss.numpy()
            actor_loss = loss_info.extra.policy_gradient_loss.numpy()
            critic_loss = loss_info.extra.value_estimation_loss.numpy()
            self.log_api('', f'loss={total_loss:<7.1f} [actor={actor_loss:<7.1f} critic={critic_loss:<7.1f}]')

            self.log_api('replay_buffer.clear', '()')
            replay_buffer.clear()

            self.on_train_iteration_end(loss=total_loss, actor_loss=actor_loss, critic_loss=critic_loss)
            if tc.training_done:
                break
        return


# noinspection PyUnresolvedReferences
class TfRandomAgent(TfAgent):
    """ creates a new random agent based on uniform random actions.

        Args:
            model_config: the model configuration including the name of the target gym environment
                as well as the neural network architecture.
    """

    def __init__(self, model_config: core.ModelConfig):
        super().__init__(model_config=model_config)
        self._set_trained_policy()

    def _set_trained_policy(self):
        """Tf-Agents Random Implementation of the train loop."""
        self.log('Creating environment...')
        train_env = self._create_env()
        action_spec = train_env.action_spec()
        timestep_spec = train_env.time_step_spec()

        self.log_api('RandomTFPolicy', 'create')
        self._trained_policy = random_tf_policy.RandomTFPolicy(timestep_spec, action_spec)
        self._agent_context._is_policy_trained = True

    def load_implementation(self, directory: str):
        """NoOps implementation, since we don't save/load random policies."""
        pass

    def save_implementation(self, directory: str):
        """NoOps implementation, since we don't save/load random policies."""
        pass

    # noinspection DuplicatedCode
    def train_implementation(self, train_context: core.TrainContext):
        self.log("Training...")
        train_env = self._create_env()
        while True:
            self.on_train_iteration_begin()
            # ensure that 1 episode is played during the iteration
            time_step = train_env.reset()
            while not time_step.is_last():
                action_step = self._trained_policy.action(time_step)
                time_step = train_env.step(action_step.action)
            self.on_train_iteration_end(math.nan)
            if train_context.training_done:
                break
        return


# noinspection PyUnresolvedReferences
class TfReinforceAgent(TfAgent):
    """ creates a new agent based on the Reinforce algorithm using the tfagents implementation.
        Reinforce is a vanilla policy gradient algorithm using a single neural networks to predict
        the actions.

        Args:
            model_config: the model configuration including the name of the target gym environment
                as well as the neural network architecture.
    """

    def __init__(self, model_config: core.ModelConfig):
        super().__init__(model_config=model_config)

    # noinspection DuplicatedCode
    def train_implementation(self, train_context: core.TrainContext):
        """Tf-Agents Reinforce Implementation of the train loop."""

        assert isinstance(train_context, core.EpisodesTrainContext)
        tc: core.EpisodesTrainContext = train_context
        self.log('Creating environment...')
        train_env = self._create_env(discount=tc.reward_discount_gamma)
        observation_spec = train_env.observation_spec()
        action_spec = train_env.action_spec()
        timestep_spec = train_env.time_step_spec()

        # SetUp Optimizer, Networks and PpoAgent
        self.log_api('AdamOptimizer', 'create')
        optimizer = tf.compat.v1.train.AdamOptimizer(learning_rate=tc.learning_rate)

        self.log_api('ActorDistributionNetwork', 'create')
        actor_net = actor_distribution_network.ActorDistributionNetwork(observation_spec, action_spec,
                                                                        fc_layer_params=self.model_config.fc_layers)

        self.log_api('ReinforceAgent', 'create')
        tf_agent = reinforce_agent.ReinforceAgent(timestep_spec, action_spec, actor_network=actor_net,
                                                  optimizer=optimizer)

        self.log_api('tf_agent.initialize()')
        tf_agent.initialize()
        self._trained_policy = tf_agent.policy

        # SetUp Data collection & Buffering
        collect_data_spec = tf_agent.collect_data_spec
        self.log_api('TFUniformReplayBuffer', 'create')
        replay_buffer = TFUniformReplayBuffer(collect_data_spec, batch_size=1, max_length=tc.max_steps_in_buffer)
        self.log_api('DynamicEpisodeDriver', 'create')
        collect_driver = DynamicEpisodeDriver(train_env, tf_agent.collect_policy,
                                              observers=[replay_buffer.add_batch],
                                              num_episodes=tc.num_episodes_per_iteration)

        # Train
        collect_driver.run = common.function(collect_driver.run, autograph=False)
        tf_agent.train = common.function(tf_agent.train, autograph=False)

        self.log('Starting training...')
        while True:
            self.on_train_iteration_begin()
            msg = f'iteration {tc.iterations_done_in_training:4} of {tc.num_iterations:<4}'
            self.log_api('collect_driver.run', msg)
            collect_driver.run()

            self.log_api('replay_buffer.gather_all', msg)
            trajectories = replay_buffer.gather_all()

            self.log_api('tf_agent.train', msg)
            loss_info = tf_agent.train(experience=trajectories)
            total_loss = loss_info.loss.numpy()
            self.log_api('', f'loss={total_loss:<7.1f}')

            self.log_api('replay_buffer.clear', msg)
            replay_buffer.clear()

            self.on_train_iteration_end(loss=total_loss)
            if tc.training_done:
                break
        return


# noinspection PyUnresolvedReferences
class TfSacAgent(TfAgent):
    """ creates a new agent based on the SAC algorithm using the tfagents implementation.

        adapted from
            https://github.com/tensorflow/agents/blob/master/tf_agents/colabs/7_SAC_minitaur_tutorial.ipynb

        Args:
            model_config: the model configuration including the name of the target gym environment
                as well as the neural network architecture.
    """

    def __init__(self, model_config: core.ModelConfig):
        super().__init__(model_config=model_config)

    # noinspection DuplicatedCode
    def train_implementation(self, train_context: core.TrainContext):
        """Tf-Agents Ppo Implementation of the train loop."""

        assert isinstance(train_context, core.StepsTrainContext)
        tc: core.StepsTrainContext = train_context
        train_env = self._create_env(discount=tc.reward_discount_gamma)
        observation_spec = train_env.observation_spec()
        action_spec = train_env.action_spec()
        timestep_spec = train_env.time_step_spec()

        self.log_api('CriticNetwork',
                     f'(observation_spec, action_spec), observation_fc_layer_params=None, ' +
                     f'action_fc_layer_params=None, joint_fc_layer_params={self.model_config.fc_layers})')
        critic_net = critic_network.CriticNetwork((observation_spec, action_spec),
                                                  observation_fc_layer_params=None, action_fc_layer_params=None,
                                                  joint_fc_layer_params=self.model_config.fc_layers)

        def normal_projection_net(action_spec_arg, init_means_output_factor=0.1):
            return normal_projection_network.NormalProjectionNetwork(action_spec_arg,
                                                                     mean_transform=None,
                                                                     state_dependent_std=True,
                                                                     init_means_output_factor=init_means_output_factor,
                                                                     std_transform=sac_agent.std_clip_transform,
                                                                     scale_distribution=True)

        self.log_api('ActorDistributionNetwork',
                     f'observation_spec, action_spec, fc_layer_params={self.model_config.fc_layers}), ' +
                     f'continuous_projection_net=...)')
        actor_net = actor_distribution_network.ActorDistributionNetwork(observation_spec, action_spec,
                                                                        fc_layer_params=self.model_config.fc_layers,
                                                                        continuous_projection_net=normal_projection_net)
        # self.log_api('tf.compat.v1.train.get_or_create_global_step','()')
        # global_step = tf.compat.v1.train.get_or_create_global_step()
        self.log_api('SacAgent', f'(timestep_spec, action_spec, actor_network=..., critic_network=..., ' +
                     f'actor_optimizer=AdamOptimizer(learning_rate={tc.learning_rate}), ' +
                     f'critic_optimizer=AdamOptimizer(learning_rate={tc.learning_rate}), ' +
                     f'alpha_optimizer=AdamOptimizer(learning_rate={tc.learning_rate}), ' +
                     f'gamma={tc.reward_discount_gamma})')
        tf_agent = sac_agent.SacAgent(
            timestep_spec,
            action_spec,
            actor_network=actor_net,
            critic_network=critic_net,
            actor_optimizer=tf.compat.v1.train.AdamOptimizer(learning_rate=tc.learning_rate),
            critic_optimizer=tf.compat.v1.train.AdamOptimizer(learning_rate=tc.learning_rate),
            alpha_optimizer=tf.compat.v1.train.AdamOptimizer(learning_rate=tc.learning_rate),
            # target_update_tau=0.005,
            # target_update_period=1,
            # td_errors_loss_fn=tf.compat.v1.losses.mean_squared_error,
            gamma=tc.reward_discount_gamma)
        # reward_scale_factor=1.0,
        # gradient_clipping=None,
        # train_step_counter=global_step)
        self.log_api('tf_agent.initialize', '()')
        tf_agent.initialize()

        self._trained_policy = greedy_policy.GreedyPolicy(tf_agent.policy)
        collect_policy = tf_agent.collect_policy

        # setup and preload replay buffer
        self.log_api('TFUniformReplayBuffer', f'(data_spec=tf_agent.collect_data_spec, ' +
                     f'batch_size={train_env.batch_size}, max_length={tc.max_steps_in_buffer})')
        replay_buffer = tf_uniform_replay_buffer.TFUniformReplayBuffer(data_spec=tf_agent.collect_data_spec,
                                                                       batch_size=train_env.batch_size,
                                                                       max_length=tc.max_steps_in_buffer)

        self.log_api('DynamicStepDriver', f'(env, collect_policy, observers=[replay_buffer.add_batch], ' +
                     f'num_steps={tc.num_steps_buffer_preload})')
        initial_collect_driver = dynamic_step_driver.DynamicStepDriver(train_env,
                                                                       collect_policy,
                                                                       observers=[replay_buffer.add_batch],
                                                                       num_steps=tc.num_steps_buffer_preload)
        self.log_api('initial_collect_driver.run()')
        initial_collect_driver.run()

        # Dataset generates trajectories with shape [Bx2x...]
        dataset = replay_buffer.as_dataset(num_parallel_calls=3, sample_batch_size=tc.num_steps_sampled_from_buffer,
                                           num_steps=2).prefetch(3)
        iterator = iter(dataset)

        self.log_api('DynamicStepDriver', f'(env, collect_policy, observers=[replay_buffer.add_batch], ' +
                     f'num_steps={tc.num_steps_per_iteration})')
        collect_driver = dynamic_step_driver.DynamicStepDriver(train_env,
                                                               collect_policy,
                                                               observers=[replay_buffer.add_batch],
                                                               num_steps=tc.num_steps_per_iteration)

        # (Optional) Optimize by wrapping some of the code in a graph using TF function.
        tf_agent.train = common.function(tf_agent.train)
        collect_driver.run = common.function(collect_driver.run)

        self.log_api('for each iteration')
        self.log_api('  collect_driver.run', '()')
        self.log_api('  tf_agent.train', '(experience=...)')
        while True:
            self.on_train_iteration_begin()
            # Collect a few steps using collect_policy and save to the replay buffer.
            for _ in range(tc.num_steps_per_iteration):
                collect_driver.run()

            # Sample a batch of data from the buffer and update the agent's network.
            experience, _ = next(iterator)
            loss_info = tf_agent.train(experience)
            total_loss = loss_info.loss.numpy()
            actor_loss = loss_info.extra.actor_loss
            alpha_loss = loss_info.extra.alpha_loss
            critic_loss = loss_info.extra.critic_loss
            self.on_train_iteration_end(loss=total_loss, actor_loss=actor_loss, critic_loss=critic_loss,
                                        alpha_loss=alpha_loss)
            if tc.training_done:
                break
        return


class TfAgentAgentFactory(bcore.BackendAgentFactory):
    """Backend for TfAgents.

        Serves as a factory to create algorithm specific wrappers for the TfAgents implementations.
    """

    backend_name: str = 'tfagents'

    def get_algorithms(self) -> Dict[Type, Type[fastrlapi.backends.core.BackendAgent]]:
        """Yields a mapping of EasyAgent types to the implementations provided by this backend."""
        return {fastrlapi.agents.DqnAgent: TfDqnAgent,
                fastrlapi.agents.PpoAgent: TfPpoAgent,
                fastrlapi.agents.RandomAgent: TfRandomAgent,
                fastrlapi.agents.ReinforceAgent: TfReinforceAgent,
                fastrlapi.agents.SacAgent: TfSacAgent}



import unittest

import pytest

from fastrlapi import core, env
from fastrlapi.backends import core as bcore
from fastrlapi.callbacks import log

import fastrlapi

fastrlapi.agents.seed = 0
_lineworld_name = env._LineWorldEnv.register_with_gym()
_cartpole_name = 'CartPole-v0'


class TensorforceAgentTest(unittest.TestCase):

    def setUp(self):
        self.env_name = env._StepCountEnv.register_with_gym()

    @pytest.mark.skipif(fastrlapi.backends.core._tf_eager_execution_active, reason="_tf_eager_execution_active")
    @pytest.mark.tforce
    def test_dqn_train(self):
        from fastrlapi.backends import tforce

        model_config = core.ModelConfig(_cartpole_name, fc_layers=(100, 100))
        tc: core.StepsTrainContext = core.StepsTrainContext()
        tc.num_iterations = 10000
        tc.num_steps_buffer_preload = 500
        tc.num_iterations_between_eval = 500
        tc.max_steps_per_episode = 200
        dqn_agent = tforce.TforceDqnAgent(model_config=model_config)
        dqn_agent.train(train_context=tc, callbacks=[log.Iteration(eval_only=True), log.Agent()])
        (min_r, avg_r, max_r) = tc.eval_rewards[tc.episodes_done_in_training]
        assert avg_r > 50

    @pytest.mark.skipif(fastrlapi.backends.core._tf_eager_execution_active, reason="_tf_eager_execution_active")
    @pytest.mark.tforce
    def test_dueling_dqn_train(self):
        from fastrlapi.backends import tforce

        model_config = core.ModelConfig(_cartpole_name, fc_layers=(100,))
        tc: core.StepsTrainContext = core.StepsTrainContext()
        tc.num_iterations = 2000
        tc.num_steps_buffer_preload = 100
        tc.num_iterations_between_eval = 100
        tc.max_steps_per_episode = 200
        dqn_agent = tforce.TforceDuelingDqnAgent(model_config=model_config)
        dqn_agent.train(train_context=tc, callbacks=[log.Iteration(eval_only=True), log.Agent()])

    @pytest.mark.skipif(fastrlapi.backends.core._tf_eager_execution_active, reason="_tf_eager_execution_active")
    @pytest.mark.tforce
    def test_ppo_train(self):
        from fastrlapi.backends import tforce

        model_config = core.ModelConfig(_cartpole_name)
        tc = core.PpoTrainContext()
        tc.num_iterations = 20
        ppo_agent = tforce.TforcePpoAgent(model_config=model_config)
        ppo_agent.train(train_context=tc, callbacks=[log.Iteration(), log.Agent()])
        (min_r, avg_r, max_r) = tc.eval_rewards[tc.episodes_done_in_training]
        assert avg_r > 100

    @pytest.mark.skipif(fastrlapi.backends.core._tf_eager_execution_active, reason="_tf_eager_execution_active")
    @pytest.mark.tforce
    def test_random_train(self):
        from fastrlapi.backends import tforce

        model_config = core.ModelConfig(_cartpole_name)
        tc = core.TrainContext()
        tc.num_iterations = 50
        random_agent = tforce.TforceRandomAgent(model_config=model_config)
        random_agent.train(train_context=tc, callbacks=[log.Iteration(), log.Agent()])
        (min_r, avg_r, max_r) = tc.eval_rewards[tc.episodes_done_in_training]
        assert avg_r < 50

    @pytest.mark.skipif(fastrlapi.backends.core._tf_eager_execution_active, reason="_tf_eager_execution_active")
    @pytest.mark.tforce
    def test_reinforce_train(self):
        from fastrlapi.backends import tforce

        model_config = core.ModelConfig(_cartpole_name)
        tc = core.EpisodesTrainContext()
        tc.num_iterations = 50
        reinforce_agent = tforce.TforceReinforceAgent(model_config=model_config)
        reinforce_agent.train(train_context=tc, callbacks=[log.Iteration(), log.Agent()])
        (min_r, avg_r, max_r) = tc.eval_rewards[tc.episodes_done_in_training]
        assert avg_r > 100

    @pytest.mark.skipif(fastrlapi.backends.core._tf_eager_execution_active, reason="_tf_eager_execution_active")
    @pytest.mark.tforce
    def test_save_(self):
        from fastrlapi.backends import tforce

        model_config = core.ModelConfig(_cartpole_name)
        tc = core.PpoTrainContext()
        tc.num_iterations = 3
        ppo_agent = tforce.TforcePpoAgent(model_config=model_config)
        ppo_agent.train(train_context=tc, callbacks=[log.Iteration(), log.Agent()])
        tempdir = bcore._get_temp_path()
        bcore._mkdir(tempdir)
        ppo_agent.save(tempdir, [])

        loaded_agent = tforce.TforcePpoAgent(model_config=model_config)
        loaded_agent.load(tempdir, [])
        bcore._rmpath(tempdir)


if __name__ == '__main__':
    unittest.main()



import os
import sys
from typing import List, Tuple

import fastrlapi.core as core
import fastrlapi.backends.core as bcore


class _SaveCallback(core.AgentCallback):
    """Base class for all agent savent callbacks.

        Attributes:
            directory: the absolute path of the directory containing the persisted policies.
            saved_agents: list of tuples (episode, avg_rewards, directory) for each saved agent
    """

    def __init__(self, directory: str = None):
        """Saves the best policies (along with the agent definition) in directory.
        If directory is None the policies are written in a temp directory.

        Args:
            directory: the directory to save to, if None a temp directory is created.
        """
        directory = directory if directory else bcore._get_temp_path()
        self.directory: str = bcore._mkdir(directory)
        self.saved_agents: List[Tuple[int, float, str]] = []

    def __str__(self):
        return self.directory

    def _save(self, agent_context: core.AgentContext):
        """Saves the current policy in directory."""
        assert agent_context
        assert agent_context.train, "TrainContext not set."

        tc = agent_context.train
        min_rewards, avg_reward, max_rewards = tc.eval_rewards[tc.episodes_done_in_training]
        current_dir = f'episode_{tc.episodes_done_in_training}-avg_reward_{avg_reward}'
        current_dir = os.path.join(self.directory, current_dir)
        agent_context._agent_saver(directory=current_dir)
        self.saved_agents.append((tc.episodes_done_in_training, avg_reward, current_dir))


class Best(_SaveCallback):
    """After each policy evaluation the policy is saved if average reward is larger than all previous average
        rewards. The policies can then be loaded using agents.load()

        Attributes:
            directory: the absolute path of the directory containing the persisted policies.
            saved_agents: list of tuples (episode, avg_rewards, directory) for each saved agent
    """

    def __init__(self, directory: str = None):
        """Saves the best policies (along with the agent definition) in directory.
        If directory is None the policies are written in a temp directory.

        Args:
            directory: the directory to save to, if None a temp directory is created.
        """
        super().__init__(directory=directory)
        self._best_avg_reward = -sys.float_info.max

    def on_play_end(self, agent_context: core.AgentContext):
        if agent_context.is_eval:
            tc = agent_context.train
            min_rewards, avg_reward, max_rewards = tc.eval_rewards[tc.episodes_done_in_training]
            if avg_reward > self._best_avg_reward:
                self._save(agent_context=agent_context)
                self._best_avg_reward = avg_reward


class Every(_SaveCallback):
    """Saves the current policy every n evaluations.

        Attributes:
            directory: the absolute path of the directory containing the persisted policies.
            saved_agents: list of tuples (episode, avg_rewards, directory) for each saved agent
    """

    def __init__(self, num_evals_between_save: int = 1, directory: str = None):
        """Saves the current policy every n evaluations. In terms of episodes: the policy is saved every

        num_eval_between_saves * num_iterations_between_eval * num_episodes_per_iteration

        Args:
            num_evals_between_save: the number of evaluations between saves.
        """
        assert num_evals_between_save > 0
        super().__init__(directory=directory)
        self.num_evals_between_save: int = num_evals_between_save

    def on_play_end(self, agent_context: core.AgentContext):
        if agent_context.is_eval:
            tc = agent_context.train
            if tc.num_iterations % tc.num_iterations_between_eval == 0:
                evals = tc.num_iterations / tc.num_iterations_between_eval
                if evals % self.num_evals_between_save == 0:
                    self._save(agent_context=agent_context)



"""This module contains the core datastructures shared between fronten and backend like the definition
    of all callbacks and agent configurations.
"""

from abc import ABC
from typing import Callable, Optional, Dict, Tuple, List, Union
from enum import Flag, auto
import math

import fastrlapi.env
import fastrlapi.backends.monitor
import gym.core
import matplotlib.pyplot as plt


class GymContext(object):
    """Contains the context for gym api calls (wrapping a gym env instance)."""

    def __init__(self):
        self._monitor_env: Optional[fastrlapi.backends.monitor._MonitorEnv] = None
        self._totals = None

    def __str__(self):
        return f'MonitorEnv={self._monitor_env} Totals={self._totals}'

    @property
    def gym_env(self) -> Optional[gym.core.Env]:
        result = None
        if self._monitor_env:
            result = self._monitor_env.env
        return result


class PlotType(Flag):
    """Defines the point in time when a plot is created / updated.

    NONE: No plot is updated.
    PLAY_EPISODE: Called after the last step of each played episode. The gym environment is still
        accessible through agent_context.play-gym_env.
    PLAY_STEP: Called after each play step. The gym environment is still
        accessible through agent_context.play-gym_env.
    TRAIN_EVAL: Called after the last step of the last evaluation episode during training.
        The gym environment is accessible through agent_context.play.gym_env.
    TRAIN_ITERATION: Called after each train iteration. No gym environment is available.
    """
    NONE = 0
    PLAY_EPISODE = auto()
    PLAY_STEP = auto()
    TRAIN_EVAL = auto()
    TRAIN_ITERATION = auto()


class PyPlotContext(object):
    """Contain the context for the maplotlib.pyplot figure plotting.

    Attributes
        figure: the figure to plot to
        figsize: figure (width,height) in inches for the figure to be created.
        is_jupyter_active: True if we plot to jupyter notebook cell, False otherwise.
        max_columns: the max number of subplot columns in the pyplot figure
    """

    def __init__(self):
        self._created_subplots = PlotType.NONE
        self.figure: Optional[plt.Figure] = None
        self.figsize: (float, float) = (17, 6)
        self._call_jupyter_display = False
        self.is_jupyter_active = False
        self.max_columns = 3

    def __str__(self):
        figure_number = None
        figure_axes_len = 0
        if self.figure:
            figure_number = self.figure.number
            if self.figure.axes:
                figure_axes_len = len(self.figure.axes)
        return f'is_jupyter_active={self.is_jupyter_active} max_columns={self.max_columns} ' + \
               f'_created_subplots={self._created_subplots} figure={figure_number} axes={figure_axes_len} '

    def _is_subplot_created(self, plot_type: PlotType):
        """Yields true if a subplot of type plot_type was created by a plot callback."""
        result = ((self._created_subplots & plot_type) != PlotType.NONE)
        return result


class ModelConfig(object):
    """The model configurations, containing the name of the gym environment and the neural network architecture.

        Attributes:
            original_env_name: the name of the underlying gym environment, eg 'CartPole-v0'
            gym_env_name: the name of the actual gym environment used (a wrapper around the environment given
                by original_env_name)
            fc_layers: int tuple defining the number and size of each fully connected layer.
            seed: the seed to be used for example for the gym_env or None for no seed
    """

    _KEY_SEED = 'seed'
    _KEY_GYM_ENV = 'gym_env'
    _KEY_FC_LAYERS = 'fc_layers'

    def __init__(self,
                 gym_env_name: str,
                 fc_layers: Union[Tuple[int, ...], int, None] = None,
                 seed: Optional[int] = None):
        """
            Args:
                gym_env_name: the name of the registered gym environment to use, eg 'CartPole-v0'
                fc_layers: int tuple defining the number and size of each fully connected layer.
        """
        if fc_layers is None:
            fc_layers = (100, 100)
        if isinstance(fc_layers, int):
            fc_layers = (fc_layers,)

        assert isinstance(gym_env_name, str), "passed gym_env_name not a string."
        assert gym_env_name != "", "gym environment name is empty."
        assert fastrlapi.env._is_registered_with_gym(gym_env_name), \
            f'"{gym_env_name}" is not the name of an environment registered with OpenAI gym.' + \
            'Consider using fastrlapi.env.register_with_gym to register your environment.'
        assert fc_layers is not None, "fc_layers not set"
        assert isinstance(fc_layers, tuple), "fc_layers not a tuple"
        assert fc_layers, "fc_layers must contain at least 1 int"
        for i in fc_layers:
            assert isinstance(i, int) and i >= 1, f'{i} is not a valid size for a hidden layer'

        self.original_env_name = gym_env_name
        self.gym_env_name = None
        self.fc_layers = fc_layers
        self.seed = seed

    def __str__(self):
        return f'fc_layers={self.fc_layers} seed={self.seed} gym_env_name={self.gym_env_name}'

    @staticmethod
    def _from_dict(from_dict: Dict[str, object]):
        """Creates a new instance of ModelConfig based on the parameters contained in dict

        Returns:
            new instance of ModelConfig configured by dict
        """
        assert from_dict
        # noinspection PyTypeChecker
        fc_layers = tuple(from_dict[ModelConfig._KEY_FC_LAYERS])
        # noinspection PyTypeChecker
        result = ModelConfig(gym_env_name=str(from_dict[ModelConfig._KEY_GYM_ENV]), fc_layers=fc_layers,
                             seed=from_dict[ModelConfig._KEY_SEED])
        return result

    def _to_dict(self) -> Dict[str, object]:
        """saves this model configuration to a dict. The model_config can be recreated by a call to _from_dict

        Retunns:
            dict containing all parameters of this model_config (this does not include any policy)
        """
        result: Dict[str, object] = dict()
        result[ModelConfig._KEY_SEED] = self.seed
        result[ModelConfig._KEY_GYM_ENV] = self.original_env_name
        result[ModelConfig._KEY_FC_LAYERS] = self.fc_layers
        return result


class TrainContext(object):
    """Contains the configuration of an agents train method like the number of iterations or the learning rate
        along with data gathered sofar during the training which is identical for all implementations.

        Hints:
        o TrainContext contains all the parameters needed to control the train loop.
        o Subclasses of TrainContext may contain additional Agent (but not backend) specific parameters.

        Attributes:
            num_iterations: number of times the training is repeated (with additional data), unlimited if None
            max_steps_per_episode: maximum number of steps per episode
            learning_rate: the learning rate used in the next iteration's policy training (0,1]
            reward_discount_gamma: the factor by which a reward is discounted for each step (0,1]
            max_steps_in_buffer: size of the agents buffer in steps

            training_done: if true the train loop is terminated at the end of the current iteration
            iterations_done_in_training: the number of iterations completed so far (during training)
            episodes_done_in_iteration: the number of episodes completed in the current iteration
            episodes_done_in_training: the number of episodes completed over all iterations so far.
                The episodes played for evaluation are not included in this count.
            steps_done_in_training: the number of steps taken over all iterations so far
            steps_done_in_iteration: the number of steps taken in the current iteration

            num_iterations_between_eval: number of training iterations before the current policy is evaluated.
            num_episodes_per_eval: number of episodes played to estimate the average return and steps
            eval_rewards: dict containg the rewards statistics for each policy evaluation.
                Each entry contains the tuple (min, average, max) over the sum of rewards over all episodes
                played for the current evaluation. The dict is indexed by the current_episode.
            eval_steps: dict containg the steps statistics for each policy evaluation.
                Each entry contains the tuple (min, average, max) over the number of step over all episodes
                played for the current evaluation. The dict is indexed by the current_episode.
            loss: dict containing the loss for each iteration training. The dict is indexed by the current_episode.
    """

    def __init__(self):
        self.num_iterations: Optional[int] = None
        self.max_steps_per_episode: Optional = 1000
        self.num_iterations_between_eval: int = 10
        self.num_episodes_per_eval: int = 10
        self.learning_rate: float = 0.001
        self.reward_discount_gamma: float = 1.0
        self.max_steps_in_buffer: int = 100000

        self.training_done: bool
        self.iterations_done_in_training: int
        self.episodes_done_in_iteration: int
        self.episodes_done_in_training: int
        self.steps_done_in_training: int
        self.steps_done_in_iteration = 0
        self.loss: Dict[int, float]
        self.eval_rewards: Dict[int, Tuple[float, float, float]]
        self.eval_steps: Dict[int, Tuple[float, float, float]]
        self._reset()

    def __str__(self):
        return f'training_done={self.training_done} ' + \
               f'#iterations_done_in_training={self.iterations_done_in_training} ' + \
               f'#episodes_done_in_iteration={self.episodes_done_in_iteration} ' + \
               f'#steps_done_in_iteration={self.steps_done_in_iteration} ' + \
               f'#iterations={self.num_iterations} ' + \
               f'#max_steps_per_episode={self.max_steps_per_episode} ' + \
               f'#iterations_between_eval={self.num_iterations_between_eval} ' + \
               f'#episodes_per_eval={self.num_episodes_per_eval} ' + \
               f'#learning_rate={self.learning_rate} ' + \
               f'#reward_discount_gamma={self.reward_discount_gamma} ' + \
               f'#max_steps_in_buffer={self.max_steps_in_buffer} '

    def _reset(self):
        """Clears all values modified during a train() call."""
        self.training_done = False
        self.iterations_done_in_training = 0
        self.episodes_done_in_iteration = 0
        self.episodes_done_in_training = 0
        self.steps_done_in_training = 0
        self.steps_done_in_iteration = 0
        self.loss = dict()
        self.eval_rewards = dict()
        self.eval_steps = dict()

    def _validate(self):
        """Validates the consistency of all values, raising an exception if an inadmissible combination is detected."""
        assert self.num_iterations is None or self.num_iterations > 0, "num_iterations not admissible"
        assert self.max_steps_per_episode > 0, "max_steps_per_episode not admissible"
        assert self.num_iterations_between_eval > 0, "num_iterations_between_eval not admissible"
        assert self.num_episodes_per_eval > 0, "num_episodes_per_eval not admissible"
        assert 0 < self.learning_rate <= 1, "learning_rate not in interval (0,1]"
        assert 0 < self.reward_discount_gamma <= 1, "reward_discount_gamma not in interval (0,1]"

    @property
    def num_iterations_between_plot(self):
        """number of iterations between 2 plot updates during training.

        Returns:
            number of iterations or 0 if no plot updates should take place.
            """
        result = 0
        if self.num_iterations_between_eval:
            result = math.ceil(self.num_iterations_between_eval / 3)
        return result


class EpisodesTrainContext(TrainContext):
    """Base class for all agent which evaluate a number of episodes during each iteration:

        The train loop proceeds roughly as follows:
            for i in num_iterations
                for e in num_episodes_per_iterations
                    play episode and record steps
                train policy for num_epochs_per_iteration epochs
                if current_episode % num_iterations_between_eval == 0:
                    evaluate policy
                if training_done
                    break

        Attributes:
            num_episodes_per_iteration: number of episodes played per training iteration
            num_epochs_per_iteration: number of times the data collected for the current iteration
                is used to retrain the current policy
    """

    def __init__(self):
        self.num_episodes_per_iteration: int = 10
        self.num_epochs_per_iteration: int = 10
        super().__init__()

    def __str__(self):
        return super().__str__() + \
               f'#episodes_per_iteration={self.num_episodes_per_iteration} ' + \
               f'#epochs_per_iteration={self.num_epochs_per_iteration} '

    def _validate(self):
        """Validates the consistency of all values, raising an exception if an inadmissible combination is detected."""
        super()._validate()
        assert self.num_episodes_per_iteration > 0, "num_episodes_per_iteration not admissible"
        assert self.num_epochs_per_iteration > 0, "num_epochs_per_iteration not admissible"


class CemTrainContext(EpisodesTrainContext):
    """Holds the configuration and current training state for Cross-Entropy-Methode agents.

        Attributes:
            elite_set_fraction: fraction of the elite policy set.
            num_steps_buffer_preload: number of steps performed to initially load the policy buffer
    """

    def __init__(self):
        super().__init__()
        self.num_iterations = 100
        self.num_episodes_per_iteration: int = 50
        self.elite_set_fraction: float = 0.1
        self.num_steps_buffer_preload: int = 2000

    def __str__(self):
        return super().__str__() + f'#elite_set_fraction={self.elite_set_fraction} '

    def _validate(self):
        """Validates the consistency of all values, raising an exception if an inadmissible combination is detected."""
        super()._validate()
        assert 1 >= self.elite_set_fraction > 0, "elite_set_fraction must be in interval (0,1]"


class PpoTrainContext(EpisodesTrainContext):
    """TrainContext for Actor-Critic type agents like Ppo or Sac.

    Attributes:
        actor_loss: loss observed during training of the actor network. dict is indexed by the current_episode.
        critic_loss: loss observed during training of the critic network. dict is indexed by the current_episode.
    """

    def __init__(self):
        super().__init__()
        self.actor_loss: Dict[int, float]
        self.critic_loss: Dict[int, float]

    def _reset(self):
        self.actor_loss = dict()
        self.critic_loss = dict()
        super()._reset()


class StepsTrainContext(TrainContext):
    """Base class for all agent which evaluate a number of steps during each iteration:

        The train loop proceeds roughly as follows:
            for i in num_iterations
                for s in num_steps_per_iterations
                    play episodes and record steps
                train policy for num_epochs_per_iteration epochs
                if current_episode % num_iterations_between_eval == 0:
                    evaluate policy
                if training_done
                    break

        Attributes:
            num_steps_per_iteration: number of steps played for each iteration
            num_steps_buffer_preload: number of initial collect steps to preload the buffer
            num_steps_sampled_from_buffer: the number of steps sampled from buffer for each iteration training
    """

    def __init__(self):
        super().__init__()
        self.num_iterations = 20000
        self.num_iterations_between_eval = 1000
        self.num_steps_per_iteration: int = 1
        self.num_steps_buffer_preload: int = 1000
        self.num_steps_sampled_from_buffer: int = 64
        self.max_steps_in_buffer = 100000

    def __str__(self):
        return super().__str__() + \
               f'#steps_per_iteration={self.num_steps_per_iteration} ' + \
               f'#steps_buffer_preload={self.num_steps_buffer_preload} ' + \
               f'#steps_sampled_from_buffer={self.num_steps_sampled_from_buffer} '

    def _validate(self):
        """Validates the consistency of all values, raising an exception if an inadmissible combination is detected."""
        super()._validate()
        assert self.num_steps_per_iteration > 0, "num_steps_per_iteration not admissible"


class PlayContext(object):
    """Contains the current configuration of an agents play method like the number of episodes to play
        and the max number of steps per episode.

        The EasyAgent.play() method proceeds (roughly) as follow:

        for e in num_episodes
            play (while steps_done_in_episode < max_steps_per_episode)
            if playing_done
                break

        Attributes:
            num_episodes: number of episodes to play, unlimited if None
            max_steps_per_episode: maximum number of steps per episode, unlimited if None
            play_done: if true the play loop is terminated at the end of the current episode
            episodes_done: the number of episodes played (including the current episode).
            steps_done_in_episode: the number of steps taken in the current episode.
            steps_done: the number of steps played (over all episodes so far)

            actions: dict containing for each episode the actions taken in each step
            rewards: dict containing for each episode the rewards received in each step
            sum_of_rewards: dict containing for each episode the sum of rewards over all steps
            gym_env: the gym environment used to play
    """

    def __init__(self, train_context: Optional[TrainContext] = None):
        """
        Args:
             train_context: if set num_episodes, max_steps_per_episode and seed are set from train_context
        """
        self.num_episodes: Optional[int] = None
        self.max_steps_per_episode: Optional[int] = None

        if train_context is not None:
            self.num_episodes = train_context.num_episodes_per_eval
            self.max_steps_per_episode = train_context.max_steps_per_episode

        self.play_done: bool
        self.episodes_done: int
        self.steps_done_in_episode: int
        self.steps_done: int
        self.actions: Dict[int, List[object]]
        self.rewards: Dict[int, List[float]]
        self.sum_of_rewards: Dict[int, float]
        self.gym_env: Optional[gym.core.Env]
        self._reset()

    def __str__(self):
        return f'#episodes={self.num_episodes} ' + \
               f'max_steps_per_episode={self.max_steps_per_episode} ' + \
               f'play_done={self.play_done} ' + \
               f'episodes_done={self.episodes_done} ' + \
               f'steps_done_in_episode={self.steps_done_in_episode} ' + \
               f'steps_done={self.steps_done} '

    def _reset(self):
        """Clears all values modified during a train() call."""
        self.play_done: bool = False
        self.episodes_done: int = 0
        self.steps_done_in_episode: int = 0
        self.steps_done: int = 0
        self.actions: Dict[int, List[object]] = dict()
        self.rewards: Dict[int, List[float]] = dict()
        self.sum_of_rewards: Dict[int, float] = dict()
        self.gym_env: Optional[gym.core.Env] = None

    def _validate(self):
        """Validates the consistency of all values, raising an exception if an inadmissible combination is detected."""
        assert (self.num_episodes is None) or (self.num_episodes > 0), "num_episodes not admissible"
        assert (self.max_steps_per_episode is None) or self.max_steps_per_episode > 0, \
            "max_steps_per_episode not admissible"


class AgentContext(object):
    """Collection of state and configuration settings for a EasyAgent instance.

    Attributes:
        model: model configuration including the name of the underlying gym_environment and the
            policy's neural network archtitecture.
        train: training configuration and current train state. None if not inside a train call.
        play: play / eval configuration and current state. None if not inside a play call (directly or
            due to a evaluation inside a train loop)
        gym: context for gym environment related calls.
        pyplot: the context containing the matplotlib.pyplot figure to plot to during training or playing
    """

    def __init__(self, model: ModelConfig):
        """
        Args:
            model: model configuration including the name of the underlying gym_environment and the
                policy's neural network archtitecture.
        """
        assert isinstance(model, ModelConfig), "model not set"

        self.model: ModelConfig = model
        self.train: Optional[TrainContext] = None
        self.play: Optional[PlayContext] = None
        self.gym: GymContext = GymContext()
        self.pyplot: PyPlotContext = PyPlotContext()
        self._is_policy_trained = False
        self._agent_saver: Optional[
            Callable[[Optional[str], Union[List[AgentCallback], AgentCallback, None]], str]] = None

    def __str__(self):
        result = f'agent_context:'
        result += f'\napi   =[{self.gym}]'
        if self.train is not None:
            result += f'\ntrain =[{self.train}] '
        if self.play is not None:
            result += f'\nplay  =[{self.play}] '
        if self.pyplot is not None:
            result += f'\npyplot=[{self.pyplot}] '
        result += f'\nmodel =[{self.model}] '
        return result

    @property
    def is_eval(self) -> bool:
        """Yields true if a policy evaluation inside an agent.train(...) call is in progress."""
        return (self.play is not None) and (self.train is not None)

    @property
    def is_play(self) -> bool:
        """Yields true if an agent.play(...) call is in progress, but not a policy evaluation"""
        return (self.play is not None) and (self.train is None)

    def _is_plot_ready(self, plot_type: PlotType) -> bool:
        """Yields true if any of the plots in plot_type is ready to be plotted.

        A plot_type is ready if a plot callback was registered for this type (like TRAIN_EVAL),
        the agent is in runtime state corresponding to the plot type (like in training and at the end of
        an evaluation period) and any frequency condition is met (like num_episodes_between_plot)
        """
        result = False
        if (plot_type & PlotType.PLAY_EPISODE) != PlotType.NONE:
            result = result | (self.is_play and self.pyplot._is_subplot_created(PlotType.PLAY_EPISODE))
        if (plot_type & PlotType.PLAY_STEP) != PlotType.NONE:
            result = result | (self.is_play and self.pyplot._is_subplot_created(PlotType.PLAY_STEP))
        if (plot_type & PlotType.TRAIN_EVAL) != PlotType.NONE:
            train_result = self.is_eval
            train_result = train_result and self.pyplot._is_subplot_created(PlotType.TRAIN_EVAL)
            train_result = train_result and (self.play.episodes_done == self.train.num_episodes_per_eval)
            result = result | train_result
        if (plot_type & PlotType.TRAIN_ITERATION) != PlotType.NONE:
            train_result = self.is_train
            train_result = train_result and self.pyplot._is_subplot_created(PlotType.TRAIN_ITERATION)
            train_result = train_result and \
                           self.train.num_iterations_between_plot > 0 and \
                           ((self.train.iterations_done_in_training % self.train.num_iterations_between_plot) == 0)
            result = result | train_result
        return result

    @property
    def is_train(self) -> bool:
        """Yields true if an agent.train(...) call is in progress, but not a policy evaluation."""

        return (self.train is not None) and (self.play is None)


class AgentCallback(ABC):
    """Base class for all callbacks monitoring the backend algorithms api calls or the api calls to the gym
        environment"""

    def on_api_log(self, agent_context: AgentContext, api_target: str, log_msg: str):
        """Logs a call to the api of the agents implementation library / framework."""
        pass

    def on_log(self, agent_context: AgentContext, log_msg: str):
        """Logs a general message"""
        pass

    def on_gym_init_begin(self, agent_context: AgentContext):
        """called when the monitored environment begins the instantiation of a new gym environment.

            Args:
                agent_context: api_context passed to calling agent
        """

    def on_gym_init_end(self, agent_context: AgentContext):
        """called when the monitored environment completed the instantiation of a new gym environment.

        Args:
            agent_context: api_context passed to calling agent
        """
        pass

    def on_gym_reset_begin(self, agent_context: AgentContext, **kwargs):
        """Before a call to gym.reset

            Args:
                agent_context: api_context passed to calling agent
                kwargs: the args to be passed to the underlying environment
        """

    def on_gym_reset_end(self, agent_context: AgentContext, reset_result: Tuple, **kwargs):
        """After a call to gym.reset was completed

        Args:
            agent_context: api_context passed to calling agent
            reset_result: object returned by gym.reset
            kwargs: args passed to gym.reset
        """
        pass

    def on_gym_step_begin(self, agent_context: AgentContext, action):
        """Before a call to gym.step

        Args:
            agent_context: api_context passed to calling agent
            action: the action to be passed to the underlying environment
        """
        pass

    def on_gym_step_end(self, agent_context: AgentContext, action, step_result: Tuple):
        """After a call to gym.step was completed

        Args:
            agent_context: api_context passed to calling agent
            action: the action to be passed to the underlying environment
            step_result: (observation,reward,done,info) tuple returned by gym.step
        """
        pass

    def on_play_episode_begin(self, agent_context: AgentContext):
        """Called once at the start of new episode to be played (during play or eval, but not during train). """

    def on_play_episode_end(self, agent_context: AgentContext):
        """Called once after an episode is done or stopped (during play or eval, but not during train)."""

    def on_play_begin(self, agent_context: AgentContext):
        """Called once at the entry of an agent.play() call (during play or eval, but not during train). """

    def on_play_end(self, agent_context: AgentContext):
        """Called once before exiting an agent.play() call (during play or eval, but not during train)"""

    def on_play_step_begin(self, agent_context: AgentContext, action):
        """Called once before a new step is taken in the current episode (during play or eval, but not during train).

            Args:
                 agent_context: the context describing the agents current configuration
                 action: the action to be passed to the upcoming gym_env.step call
        """

    def on_play_step_end(self, agent_context: AgentContext, action, step_result: Tuple):
        """Called once after a step is completed in the current episode (during play or eval, but not during train)."""

    def on_train_begin(self, agent_context: AgentContext):
        """Called once at the entry of an agent.train() call. """

    def on_train_end(self, agent_context: AgentContext):
        """Called once before exiting an agent.train() call"""

    def on_train_iteration_begin(self, agent_context: AgentContext):
        """Called once at the start of a new iteration. """

    def on_train_iteration_end(self, agent_context: AgentContext):
        """Called once after the current iteration is completed"""


class _PostProcessCallback(AgentCallback):
    pass


class _PreProcessCallback(AgentCallback):
    pass



import unittest
from fastrlapi import agents
from fastrlapi.callbacks import duration, log


class DurationTest(unittest.TestCase):
    def test_single_episode(self):
        agent = agents.PpoAgent("CartPole-v0")
        count=log._CallbackCounts()
        agent.train([duration._SingleEpisode(), log._Callbacks(), count])
        assert count.train_iteration_begin_count == 1


if __name__ == '__main__':
    unittest.main()



from typing import Type, Dict

from fastrlapi.backends import core as bcore
import fastrlapi.core


class TensorforceNotActiveAgent(bcore.BackendAgent):

    def __init__(self, model_config: fastrlapi.core.ModelConfig, backend_name: str):
        raise NotImplementedError(
            "Call agents.activate_tensorforce() to activate tensorforce before instantiating the first agent."
            "This agent is implemented by tensorforce. Due to an incompatibility"
            "between tensorforce and tfagents their agents can not be instantiated in the"
            "same python runtime instance (conflicting excpectations on tensorflows eager execution mode)."
        )


class TfAgentsNotActiveAgent(bcore.BackendAgent):

    def __init__(self, model_config: fastrlapi.core.ModelConfig, backend_name: str):
        raise NotImplementedError(
            "Do not call agents.activate_tensorforce() before instantiating a tfagents based agent."
            "This agent is implemented by tfagents. Due to an incompatibility"
            "between tensorforce and tfagents their agents can not be instantiated in the"
            "same python runtime instance (conflicting excpectations on tensorflows eager execution mode)."
        )


class SetTensorforceBackendAgent(bcore.BackendAgent):

    def __init__(self, model_config: fastrlapi.core.ModelConfig, backend_name: str):
        raise NotImplementedError(
            "Set the backend='tensorforce' argument in the fastrlapi constructor call. "
            "This agents default implementation is implemented by tfagents. Due to an incompatibility"
            "between tensorforce and tfagents their agents can not be instantiated in the"
            "same python runtime instance (conflicting excpectations on tensorflows eager execution mode)."
        )


class NotImplementedYetAgent(bcore.BackendAgent):

    def __init__(self, model_config: fastrlapi.core.ModelConfig, backend_name: str):
        raise NotImplementedError("fastrlapi implementation is pending.")


class DefaultAgentFactory(bcore.BackendAgentFactory):
    """Backend which redirects all calls to the some default implementation."""

    def __init__(self, register_tensorforce: bool):
        self.register_tensorforce = register_tensorforce

    backend_name = 'default'

    def get_algorithms(self) -> Dict[Type, Type[fastrlapi.backends.core.BackendAgent]]:
        """Yields a mapping of EasyAgent types to the implementations provided by this backend."""
        #                fastrlapi.agents.CemAgent: fastrlapi.backends.kerasrl.KerasRlCemAgent,
        #                fastrlapi.agents.DoubleDqnAgent: fastrlapi.backends.kerasrl.KerasRlDoubleDqnAgent,
        if self.register_tensorforce:
            import fastrlapi.backends.tforce

            result = {
                fastrlapi.agents.DqnAgent: SetTensorforceBackendAgent,
                fastrlapi.agents.DoubleDqnAgent: TfAgentsNotActiveAgent,
                fastrlapi.agents.DuelingDqnAgent: fastrlapi.backends.tforce.TforceDuelingDqnAgent,
                fastrlapi.agents.PpoAgent: SetTensorforceBackendAgent,
                fastrlapi.agents.RandomAgent: SetTensorforceBackendAgent,
                fastrlapi.agents.ReinforceAgent: SetTensorforceBackendAgent,
                fastrlapi.agents.SacAgent: TfAgentsNotActiveAgent}
        else:
            import fastrlapi.backends.tfagents

            result = {
                fastrlapi.agents.DqnAgent: fastrlapi.backends.tfagents.TfDqnAgent,
                fastrlapi.agents.DoubleDqnAgent: NotImplementedYetAgent,
                fastrlapi.agents.DuelingDqnAgent: TensorforceNotActiveAgent,
                fastrlapi.agents.PpoAgent: fastrlapi.backends.tfagents.TfPpoAgent,
                fastrlapi.agents.RandomAgent: fastrlapi.backends.tfagents.TfRandomAgent,
                fastrlapi.agents.ReinforceAgent: fastrlapi.backends.tfagents.TfReinforceAgent,
                fastrlapi.agents.SacAgent: fastrlapi.backends.tfagents.TfSacAgent}
        return result



# __init__.py

import fastrlapi.agents
import fastrlapi.callbacks.duration
import fastrlapi.callbacks.log
import fastrlapi.callbacks.plot
import fastrlapi.callbacks.save
import fastrlapi.core
import fastrlapi.env

# Version of the easy_agents package
__version__ = "1.5.0"


